"""
Multiprocessing functions
"""
# Info
__author__ = 'Jason Anthony Vander Heiden'
from presto import __version__, __date__

# Imports
import ctypes
import os
import signal
import sys
import multiprocessing as mp
from collections import OrderedDict
from time import time
from Bio import SeqIO
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord

# Presto imports
from presto.Defaults import default_coord, default_delimiter, default_out_args
from presto.Annotation import getCoordKey
from presto.IO import getFileType, readSeqFile, countSeqFile, countSeqSets, \
                      getOutputHandle, printLog, printProgress, printWarning, printError

# Constants
TERMINATION_SENTINEL = None
EXCEPTION_SENTINEL = None


class SeqData:
    """
    Class defining sequence data objects for worker processes

    Attributes:
      id : unique identifier
      data : single object or a list of data objects.
      valid : if True data is suitable for processing.
    """
    # Instantiation
    def __init__(self, id, data):
        """
        Initializer

        Arguments:
          id :  unique identifier.
          data : single object or a list of data objects.

        Returns:
          presto.Multiprocessing.SeqData
        """
        self.id = id
        self.data = data
        self.valid = (id is not None and data is not None)

    # Set boolean evaluation to valid value
    def __bool__(self):
        """
        Boolean evaluation

        Returns:
          bool : True if the valid attribute is True
        """
        return self.valid

    # Set length evaluation to number of data records
    def __len__(self):
        """
        Length evaluation

        Returns:
          int : number of objects in the data attribute.
        """
        if isinstance(self.data, SeqRecord) or isinstance(self.data, Seq):
            return 1
        elif self.data is None:
            return 0
        else:
            return len(self.data)


class SeqResult:
    """
    Class defining sequence result objects for collector processes

    Attributes:
      id : unique identifier
      data : single unprocessed object or a list of unprocessed data objects.
      results : single processed object or a list of processed data objects.
      valid : if True processing was successful.
      log : dictionary containing the processing log.
    """
    def __init__(self, id, data):
        """
        Initializer

        Arguments:
          id :  unique identifier.
          data : single unprocessed object or a list of unprocessed data objects.

        Returns:
          presto.Multiprocessing.SeqResult
        """
        self.id = id
        self.data = data
        self.results = None
        self.valid = False
        self.log = OrderedDict([('ID', id)])

    def __bool__(self):
        """
        Boolean evaluation

        Returns:
          bool : True if the valid attribute is True.
        """
        return self.valid

    def __len__(self):
        """
        Length evaluation

        Returns:
          int : number of objects in the results attribute.
        """
        if isinstance(self.results, SeqRecord) or isinstance(self.results, Seq):
            return 1
        elif self.results is None:
            return 0
        else:
            return len(self.results)

    @property
    def data_count(self):
        """
        Data length

        Returns:
          int : number of objects in the data attribute.
        """
        if isinstance(self.data, SeqRecord) or isinstance(self.data, Seq):
            return 1
        elif self.data is None:
            return 0
        else:
            return len(self.data)


def manageProcesses(feed_func, work_func, collect_func,
                    feed_args={}, work_args={}, collect_args={},
                    nproc=None, queue_size=None):
    """
    Manages feeder, worker and collector processes

    Arguments:
      feed_func (function): Data Queue feeder function.
      work_func (function): Worker function.
      collect_func (function): Result Queue collector function.
      feed_args (dict): Dictionary of arguments to pass to feed_func.
      work_args (dict): Dictionary of arguments to pass to work_func.
      collect_args (dict): Dictionary of arguments to pass to collect_func.
      nproc (int): Number of processQueue processes;
                   if None defaults to the number of CPUs
      queue_size (int): Maximum size of the argument queue;
                        if None defaults to 2*nproc

    Returns:
      dict: Dictionary of collector results
    """
    # Define signal handler that raises KeyboardInterrupt
    def _signalHandler(s, f):
        raise SystemExit

    # Define function to terminate child processes
    def _terminate():
        sys.stderr.write('NOTICE> Terminating child processes...  ')
        # Terminate feeders
        feeder.terminate()
        feeder.join()
        # Terminate workers
        for w in workers:
            w.terminate()
            w.join()
        # Terminate collector
        collector.terminate()
        collector.join
        sys.stderr.write('Done.\n')

    # Raise SystemExit upon termination signal
    signal.signal(signal.SIGTERM, _signalHandler)

    # Define number of processes and queue size
    if nproc is None:  nproc = mp.cpu_count()
    if queue_size is None:  queue_size = nproc * 2

    # Define shared child process keep alive flag
    alive = mp.Value(ctypes.c_bool, True)

    # Define shared data queues
    data_queue = mp.Queue(queue_size)
    result_queue = mp.Queue(queue_size)
    # TODO:  find out what's up with this context shenanigans
    ctx = mp.get_context()
    collect_queue = ctx.SimpleQueue()
    # Initiate manager and define shared data objects

    try:
        # Initiate feeder process
        feeder = mp.Process(target=feed_func,
                            args=(alive, data_queue),
                            kwargs=feed_args)
        feeder.start()

        # Initiate worker processes
        workers = []
        for __ in range(nproc):
            w = mp.Process(target=work_func,
                           args=(alive, data_queue, result_queue),
                           kwargs=work_args)
            w.start()
            workers.append(w)

        # Initiate collector process
        collector = mp.Process(target=collect_func,
                               args=(alive, result_queue, collect_queue),
                               kwargs=collect_args)
        collector.start()

        # Wait for feeder to finish and add sentinel objects to data_queue
        feeder.join()
        for __ in range(nproc):  data_queue.put(None)

        # Wait for worker processes to finish and add sentinel to result_queue
        for w in workers:  w.join()
        result_queue.put(None)

        # Wait for collector process to finish and add sentinel to collect_queue
        collector.join()
        collect_queue.put(None)

        # Get collector return values
        collected = collect_queue.get()
    except (KeyboardInterrupt, SystemExit):
        sys.stderr.write('NOTICE> Exit signal received.\n')
        _terminate()
        sys.exit()
    except Exception as e:
        printError('%s.' % e, exit=False)
        _terminate()
        sys.exit()
    except:
        printError('Exiting with unknown exception.', exit=False)
        _terminate()
        sys.exit()
    else:
        if not alive.value:
            printError('Exiting due to child process error.', exit=False)
            _terminate()
            sys.exit()

    return collected


def feedSeqQueue(alive, data_queue, seq_file, index_func=None, index_args={}):
    """
    Feeds the data queue with SeqRecord objects

    Arguments:
      alive : multiprocessing.Value boolean controlling whether processing
              continues; when False function returns
      data_queue : multiprocessing.Queue to hold data for processing
      seq_file : Sequence file to read input from
      index_func : Function to use to define sequence sets
                   if None do not index sets and feed individual records
      index_args : Dictionary of arguments to pass to index_func

    Returns:
      None
    """
    try:
        # Read input file and index sequence sets if required
        if index_func is None:
            seq_iter = readSeqFile(seq_file)
            data_iter = ((s.id, s) for s in seq_iter)
        else:
            seq_dict = readSeqFile(seq_file, index=True)
            index_dict = index_func(seq_dict, **index_args)
            data_iter = ((k, [seq_dict[i] for i in v]) \
                         for k, v in index_dict.items())
    except:
        alive.value = False
        raise

    try:
        # Iterate over data_iter and feed data queue
        while alive.value:
            # Get data from queue
            if data_queue.full():  continue
            else:  data = next(data_iter, None)
            # Exit upon reaching end of iterator
            if data is None:  break

            # Feed queue
            data_queue.put(SeqData(*data))
        else:
            sys.stderr.write('PID %s> Error in sibling process detected. Cleaning up.\n' \
                             % os.getpid())
            return None
    except:
        alive.value = False
        raise

    return None


def feedPairQueue(alive, data_queue, seq_file_1, seq_file_2,
                  coord_type=default_coord, delimiter=default_delimiter):
    """
    Feeds the data queue with sequence pairs for processQueue processes

    Arguments:
      alive : a multiprocessing.Value boolean controlling whether processing
              continues; when False function returns
      data_queue : an multiprocessing.Queue to hold data for processing
      seq_file_1 : the name of sequence file 1
      seq_file_2 : the name of sequence file 2
      coord_type : the sequence header format
      delimiter :  a tuple of delimiters for (fields, values, value lists)

    Returns:
      None
    """
    # Function to get coordinate info
    def _key_func(x):
        return getCoordKey(x, coord_type=coord_type, delimiter=delimiter)

    # Generator function to read and check files
    def _read_pairs(seq_file_1, seq_file_2):
        iter_1 = readSeqFile(seq_file_1, index=False)
        iter_2 = readSeqFile(seq_file_2, index=False)
        for seq_1, seq_2 in zip(iter_1, iter_2):
            key_1 = getCoordKey(seq_1.description, coord_type=coord_type,
                                delimiter=delimiter)
            key_2 = getCoordKey(seq_2.description, coord_type=coord_type,
                                delimiter=delimiter)
            if key_1 == key_2:
                yield (key_1, [seq_1, seq_2])
            else:
                raise Exception('Coordinates for sequences %s and %s do not match' \
                                 % (key_1, key_2))

    try:
        # Open and parse input files
        data_iter = _read_pairs(seq_file_1, seq_file_2)

        # Iterate over data_iter and feed data queue
        while alive.value:
            # Get data from queue
            if data_queue.full():  continue
            else:  data = next(data_iter, None)
            # Exit upon reaching end of iterator
            if data is None:  break

            # Feed queue
            data_queue.put(SeqData(*data))
        else:
            sys.stderr.write('PID %s> Error in sibling process detected. Cleaning up.\n' \
                             % os.getpid())
            return None
    except:
        alive.value = False
        raise

    return None


def processSeqQueue(alive, data_queue, result_queue, process_func, process_args={}):
    """
    Pulls from data queue, performs calculations, and feeds results queue

    Arguments:
      alive : multiprocessing.Value boolean controlling whether processing
              continues; when False function returns
      data_queue : multiprocessing.Queue holding data to process
      result_queue : multiprocessing.Queue to hold processed results
      process_func : function to use for processing sequences
      process_args : Dictionary of arguments to pass to process_func

    Returns:
      None
    """
    try:
        # Iterator over data queue until sentinel object reached
        while alive.value:
            # Get data from queue
            if data_queue.empty():  continue
            else:  data = data_queue.get()
            # Exit upon reaching sentinel
            if data is None:  break

            # Perform work
            result = process_func(data, **process_args)

            #import cProfile
            #prof = cProfile.Profile()
            #result = prof.runcall(process_func, data, **process_args)
            #prof.dump_stats('worker-%d.prof' % os.getpid())

            # Feed results to result queue
            result_queue.put(result)
        else:
            sys.stderr.write('PID %s> Error in sibling process detected. Cleaning up.\n' \
                             % os.getpid())
            return None
    except:
        alive.value = False
        printError('Error processing sequence with ID: %s.' % data.id, exit=False)
        raise

    return None


def collectSeqQueue(alive, result_queue, collect_queue, seq_file, label,
                    index_field=None, out_file=None, out_args=default_out_args):
    """
    Pulls from results queue, assembles results and manages log and file IO

    Arguments:
      alive : a multiprocessing.Value boolean controlling whether processing
              continues; when False function returns.
      result_queue : Multiprocessing.Queue holding worker results.
      collect_queue : Multiprocessing.Queue to store collector return values.
      seq_file : sample sequence file name.
      label : task label used to tag the output files.
      out_file : output file name. Automatically generated from the input file if None.
      out_args : Common output argument dictionary from parseCommonArgs.
      index_field : Field defining set membership for sequence sets
                    if None data queue contained individual records.

    Returns:
      None: Adds a dictionary with key value pairs to collect_queue containing
           'log' defining a log object,
           'out_files' defining the output file names
    """
    # Define output format
    out_type = getFileType(seq_file) if out_args['out_type'] is None \
               else out_args['out_type']

    # Wrapper for opening handles and writers
    def _open(x, label=label, out_file=out_file):
        if out_file is not None and x == 'pass':
            handle = open(out_file, 'w')
        else:
            handle = getOutputHandle(seq_file,
                                     out_label='%s-%s' % (label, x),
                                     out_dir=out_args['out_dir'],
                                     out_name=out_args['out_name'],
                                     out_type=out_type)
        return handle

    try:
        # Count records
        if index_field is None:
            result_count = countSeqFile(seq_file)
        else:
            result_count = countSeqSets(seq_file, index_field, out_args['delimiter'])

        # Define log handle
        if out_args['log_file'] is None:
            log_handle = None
        else:
            log_handle = open(out_args['log_file'], 'w')
    except:
        alive.value = False
        raise

    try:
        # Initialize output handles
        pass_handle, fail_handle = None, None

        # Iterator over results queue until sentinel object reached
        start_time = time()
        set_count = seq_count = pass_count = fail_count = 0
        while alive.value:
            # Get result from queue
            if result_queue.empty():  continue
            else:  result = result_queue.get()
            # Exit upon reaching sentinel
            if result is None:  break

            # Print progress for previous iteration
            printProgress(set_count, result_count, 0.05, start_time=start_time)

            # Update counts for current iteration
            set_count += 1
            seq_count += result.data_count

            # Write log
            printLog(result.log, handle=log_handle)

            # Write records
            if result:
                pass_count += 1
                try:
                    SeqIO.write(result.results, pass_handle, out_type)
                except AttributeError:
                    # Open pass file
                    pass_handle = _open('pass')
                    SeqIO.write(result.results, pass_handle, out_type)
            else:
                fail_count += 1
                if out_args['failed']:
                    try:
                        SeqIO.write(result.data, fail_handle, out_type)
                    except AttributeError:
                        # Open fail file
                        fail_handle = _open('fail')
                        SeqIO.write(result.data, fail_handle, out_type)
        else:
            sys.stderr.write('PID %s> Error in sibling process detected. Cleaning up.\n' \
                             % os.getpid())
            return None

        # Print total counts
        printProgress(set_count, result_count, 0.05, start_time=start_time)

        # Update return values
        log = OrderedDict()
        log['OUTPUT'] = os.path.basename(pass_handle.name) if pass_handle is not None else None
        log['SEQUENCES'] = seq_count
        if index_field is not None:
            log['SETS'] = set_count
        log['PASS'] = pass_count
        log['FAIL'] = fail_count

        # Close file handles and generate return data
        #collect_dict = {'log': log, 'pass': None, 'fail': None}
        collect_dict = {'log': log, 'out_files': []}
        if pass_handle is not None:
            #collect_dict['pass'] = pass_handle.name
            collect_dict['out_files'].append(pass_handle.name)
            pass_handle.close()
        if fail_handle is not None:
            #collect_dict['fail'] = fail_handle.name
            collect_dict['out_files'].append(fail_handle.name)
            fail_handle.close()
        if log_handle is not None:
            log_handle.close()
        collect_queue.put(collect_dict)
    except:
        alive.value = False
        raise

    return None


def collectPairQueue(alive, result_queue, collect_queue, seq_file_1, seq_file_2, label,
                     out_file=None, out_args=default_out_args):
    """
    Pulls from results queue, assembles results and manages log and file IO

    Arguments:
      alive : a multiprocessing.Value boolean controlling whether processing
              continues; when False function returns.
      result_queue : a multiprocessing.Queue holding worker results.
      collect_queue : a multiprocessing.Queue holding collector return values.
      seq_file_1 : the first sequence file name.
      seq_file_2 : the second sequence file name.
      label : task label used to tag the output files.
      out_file : output file name. Automatically generated from the input file if None.
      out_args : common output argument dictionary from parseCommonArgs.

    Returns:
      None: adds a dictionary of {log: log object, out_files: output file names} to collect_queue.
    """
    # Define output format
    out_type = getFileType(seq_file_1) if out_args['out_type'] is None \
        else out_args['out_type']

    # Define output names
    if out_args['out_name'] is None:
        out_name_1, out_name_2 = None, None
    else:
        out_name_1 = '%s-1' % out_args['out_name']
        out_name_2 = '%s-2' % out_args['out_name']

    # Wrapper for opening handles and writers
    def _open(x, in_file, out_name, out_file=out_file):
        if out_file is not None and x == 'pass':
            handle = open(out_file, 'w')
        else:
            handle = getOutputHandle(in_file,
                                     out_label='%s-%s' % (label, x),
                                     out_dir=out_args['out_dir'],
                                     out_name=out_name,
                                     out_type=out_type)
        return handle

    try:
        # Count input size
        result_count = countSeqFile(seq_file_1)

        # Define log handle
        if out_args['log_file'] is None:
            log_handle = None
        else:
            log_handle = open(out_args['log_file'], 'w')
    except:
        alive.value = False
        raise

    try:
        # Initialize file handles
        pass_handle, fail_handle_1, fail_handle_2 = None, None, None

        # Iterator over results queue until sentinel object reached
        start_time = time()
        iter_count = pass_count = fail_count = 0
        while alive.value:
            # Get result from queue
            if result_queue.empty():
                continue
            else:
                result = result_queue.get()
            # Exit upon reaching sentinel
            if result is None:  break

            # Print progress for previous iteration
            printProgress(iter_count, result_count, 0.05, start_time=start_time)

            # Update counts for iteration
            iter_count += 1

            # Write log
            printLog(result.log, handle=log_handle)

            # Write assembled sequences
            if result:
                pass_count += 1
                try:
                    SeqIO.write(result.results, pass_handle, out_type)
                except AttributeError:
                    # Open pass file
                    pass_handle = _open('pass', seq_file_1, out_args['out_name'])
                    SeqIO.write(result.results, pass_handle, out_type)
            else:
                fail_count += 1
                if out_args['failed']:
                    try:
                        SeqIO.write(result.data[0], fail_handle_1, out_type)
                        SeqIO.write(result.data[1], fail_handle_2, out_type)
                    except AttributeError:
                        # Open fail file
                        fail_handle_1 = _open('fail', seq_file_1, out_name_1)
                        fail_handle_2 = _open('fail', seq_file_2, out_name_2)
                        SeqIO.write(result.data[0], fail_handle_1, out_type)
                        SeqIO.write(result.data[1], fail_handle_2, out_type)

        else:
            sys.stderr.write('PID %s> Error in sibling process detected. Cleaning up.\n' \
                             % os.getpid())
            return None

        # Print total counts
        printProgress(iter_count, result_count, 0.05, start_time=start_time)

        # Update return values
        log = OrderedDict()
        log['OUTPUT'] = os.path.basename(pass_handle.name) if pass_handle is not None else None
        log['PAIRS'] = iter_count
        log['PASS'] = pass_count
        log['FAIL'] = fail_count

        # Close file handles and generate return data
        # collect_dict = {'log': log, 'pass': None, 'fail': None}
        collect_dict = {'log': log, 'out_files': []}
        if pass_handle is not None:
            # collect_dict['pass'] = pass_handle.name
            collect_dict['out_files'].append(pass_handle.name)
            pass_handle.close()
        if fail_handle_1 is not None:
            # collect_dict['fail'] = fail_handle.name
            collect_dict['out_files'].append(fail_handle_1.name)
            fail_handle_1.close()
        if fail_handle_2 is not None:
            # collect_dict['fail'] = fail_handle.name
            collect_dict['out_files'].append(fail_handle_2.name)
            fail_handle_2.close()
        if log_handle is not None:
            log_handle.close()
        collect_queue.put(collect_dict)
    except:
        alive.value = False
        raise

    return None



"""
Unit tests for MaskPrimers
"""
# Info
__author__    = 'Jason Anthony Vander Heiden'

# Imports
import os
import sys
import time
import unittest
from collections import OrderedDict
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord

# Presto imports
from presto.Sequence import getDNAScoreDict, localAlignment, scoreAlignment, extractAlignment, \
                            maskSeq, PrimerAlignment

# Paths
test_path = os.path.dirname(os.path.realpath(__file__))

# Import script
sys.path.append(os.path.join(test_path, os.pardir, 'bin'))
import MaskPrimers


class TestMaskPrimers(unittest.TestCase):
    def setUp(self):
        print('-> %s()' % self._testMethodName)
        # Test Ns
        seq_n = [Seq('CCACGTTTTAGTAATTAATA'),
                 Seq('CCNCGTTTTAGTAATTAATA'),
                 Seq('GGGCGTTTTAGTAATTAATA'),
                 Seq('GGNNGTTTTACTAATTAATA'),
                 Seq('NNGCNNNNNACTAATTAATA'),
                 Seq('GGGATANNNACTAATTAATA'),
                 Seq('NNNNNNNNNNNNNNNNNNNN')]
        self.records_n = [SeqRecord(s, id='SEQ%i' % i, name='SEQ%i' % i, description='')
                          for i, s in enumerate(seq_n, start=1)]
        self.primers_n =  OrderedDict([('PR1', 'CACGTTTT'),
                                       ('PR2', 'GGCGTTTT'),
                                       ('PR3', 'GGANATAA')])

        # (primer name, error rate)
        self.align_n = [('PR1', 0.0/8),
                        ('PR1', 1.0/8),
                        ('PR2', 0.0/8),
                        ('PR2', 2.0/8),
                        ('PR3', 3.0/8),
                        ('PR3', 3.0/8),
                        (None, 1.0)]
        # Score primers with start=1
        self.score_n = [('PR1', 0.0/8),
                        ('PR1', 1.0/8),
                        ('PR2', 0.0/8),
                        ('PR2', 2.0/8),
                        ('PR2', 6.0/8),
                        ('PR3', 3.0/8),
                        (None, 1.0)]
        # Extract primers with start=4 and length=8
        self.extract_n = ['GTTTTAGT',
                          'GTTTTAGT',
                          'GTTTTAGT',
                          'GTTTTACT',
                          'NNNNNACT',
                          'TANNNACT',
                          'NNNNNNNN']

        # Test indels
        seq_indel = [Seq('CGGATCTTCTACTCCATACGTCCGTCAGTCGTGGATCTGATCTAGCTGCGCCTTTTTCTCAG'),
                     Seq('CGGATCTTCTACTCAAAACCGTCCTCAGTCGTGGATCTGGTCTAGCTGGGGCTGTTTCCCTG'),
                     Seq('GGTTTAAGTTAAGATAATACGTCCGTCAGTCGTGATCTGTTTTACATGGGGGCATCACCCAG'),
                     Seq('CAACCACATGGGTCCTCTAGAGAATCCCCTGAGAGCTCCGTTCCTCACCATGGACTGGACCT'),
                     Seq('CAACCACATGGTCCTCTAGAGAATCCCCTGAGAGCTCCGTTCCTCACCATGGACTGGACCTG'),
                     Seq('TCTGTCCTCTAGAGAATCCCCTGAGAGCTCCGTTCCTCACCATGGACTGGACCTCAACCACA'),
                     Seq('AGGTGAAGAAGCCTGGGGCCTCCGTGAAGGTCTCCTGCTCGGCTTCTGGATACGCCTTCACC'),
                     Seq('A--TGAAGAAGCCTGGGGCCTCCGTGAAGGTCTCCTGCTCGGCTTCTGGATACGCCTTCACC'),
                     Seq('ANNTGAAGAAGNNTGGGGCCTCCGTGAAGGTCTCCTGCTCGGCTTCTGGATACGCCTTCACC'),
                     Seq('--------------------------------------------------------------')]
        self.records_indel = [SeqRecord(s, id='SEQ%i' % i, name='SEQ%i' % i, description='')
                              for i, s in enumerate(seq_indel, start=1)]
        self.primers_indel =  OrderedDict([('PR1', 'AATACGTCCGTCAGTCGTGGATGT'),
                                           ('PR2', 'CATCTTCCTCTA'),
                                           ('PR3', 'GTGAAGAGCCTG')])

        # (primer name, error rate)
        self.align_indel = [('PR1', (2.0 + 0)/24),
                            ('PR1', (4.0 + 1)/24),
                            ('PR1', (2.0 + 1)/24),
                            ('PR2', (2.0 + 1)/12),
                            ('PR2', (2.0 + 0)/12),
                            ('PR2', (0.0 + 3)/12),
                            ('PR3', (0.0 + 1)/12),
                            ('PR3', (0.0 + 2)/12),
                            ('PR3', (3.0 + 1)/12),
                            (None, 1.0)]

        # Masking test case
        mask_seq = SeqRecord(Seq('CCACGTTTTAGTAATTAATA'), id='SEQ', description='SEQ')
        self.mask_primer = PrimerAlignment(mask_seq)
        self.mask_primer.primer = 'A'
        self.mask_primer.align_seq =    'CCACGTTTT'
        self.mask_primer.align_primer = '---CGTTTT'
        self.mask_primer.start = 3
        self.mask_primer.end = 9
        self.mask_primer.gaps = 0
        self.mask_primer.error = 0
        self.mask_primer.rev_primer = False
        self.mask_primer.valid = True

        self.mask_seq_cut = SeqRecord(Seq('AGTAATTAATA'), id='SEQ|PRIMER=A|BARCODE=CCA')
        self.mask_seq_mask = SeqRecord(Seq('NNNNNNAGTAATTAATA'), id='SEQ|PRIMER=A|BARCODE=CCA')
        self.mask_seq_trim = SeqRecord(Seq('CGTTTTAGTAATTAATA'), id='SEQ|PRIMER=A|BARCODE=CCA')
        self.mask_seq_tag = SeqRecord(Seq('CCACGTTTTAGTAATTAATA'), id='SEQ|PRIMER=A|BARCODE=CCA')
        self.mask_seq_cut_barcodelen = SeqRecord(Seq('AGTAATTAATA'), id='SEQ|PRIMER=A|BARCODE=CA')

        # Start clock
        self.start = time.time()

    def tearDown(self):
        # End clock
        t = time.time() - self.start
        print('<- %s() %.3f' % (self._testMethodName, t))

    #@unittest.skip('-> extractAlignment() skipped\n')
    def test_extractAlignment(self):
        align = [extractAlignment(x, start=4, length=8)
                 for x in self.records_n]
        for x in align:
            print('  %s>' % x.seq.id)
            print('      SEQ> %s' % x.seq.seq)
            print('  ALN-SEQ> %s' % x.align_seq)
            print('   ALN-PR> %s' % x.align_primer)
            print('   PRIMER> %s' % x.primer)
            print('    START> %s' % x.start)
            print('      END> %s' % x.end)
            print('     GAPS> %i' % x.gaps)
            print('    ERROR> %f\n' % x.error)

        self.assertListEqual([x for x in self.extract_n],
                             [str(x.primer) for x in align])

    #@unittest.skip('-> scoreAlignment() skipped\n')
    def test_scoreAlignment(self):
        score_dict = getDNAScoreDict(mask_score=(0, 1), gap_score=(0, 0))
        align = [scoreAlignment(x, self.primers_n, start=1, score_dict=score_dict)
                 for x in self.records_n]
        for x in align:
            print('  %s>' % x.seq.id)
            print('      SEQ> %s' % x.seq.seq)
            print('  ALN-SEQ> %s' % x.align_seq)
            print('   ALN-PR> %s' % x.align_primer)
            print('   PRIMER> %s' % x.primer)
            print('    START> %s' % x.start)
            print('      END> %s' % x.end)
            print('     GAPS> %i' % x.gaps)
            print('    ERROR> %f\n' % x.error)

        self.assertListEqual([(x, round(y, 4)) for x, y in self.score_n],
                             [(x.primer, round(x.error, 4)) for x in align])

    #@unittest.skip('-> localAlignment() skipped\n')
    def test_localAlignment(self):
        score_dict = getDNAScoreDict(mask_score=(0, 1), gap_score=(0, 0))

        # N character tests
        print('TEST Ns>')
        align = [localAlignment(x, self.primers_n, max_error=0.2, score_dict=score_dict)
                 for x in self.records_n]
        for x in align:
            print('  %s>' % x.seq.id)
            print('      SEQ> %s' % x.seq.seq)
            print('  ALN-SEQ> %s' % x.align_seq)
            print('   ALN-PR> %s' % x.align_primer)
            print('   PRIMER> %s' % x.primer)
            print('    START> %s' % x.start)
            print('      END> %s' % x.end)
            print('     GAPS> %i' % x.gaps)
            print('    ERROR> %f\n' % x.error)

        self.assertListEqual([(x, round(y, 4)) for x, y in self.align_n],
                             [(x.primer, round(x.error, 4)) for x in align])

        # Indel tests
        print('TEST INDELS>')
        align = [localAlignment(x, self.primers_indel, max_error=0.2, gap_penalty=(1, 1))
                 for x in self.records_indel]
        for x in align:
            print('  %s>' % x.seq.id)
            print('      SEQ> %s' % x.seq.seq)
            print('  ALN-SEQ> %s' % x.align_seq)
            print('   ALN-PR> %s' % x.align_primer)
            print('   PRIMER> %s' % x.primer)
            print('    START> %s' % x.start)
            print('      END> %s' % x.end)
            print('     GAPS> %i' % x.gaps)
            print('    ERROR> %f\n' % x.error)

        self.assertListEqual([(x, round(y, 4)) for x, y in self.align_indel],
                             [(x.primer, round(x.error, 4)) for x in align])

    #@unittest.skip('-> maskSeq() skipped\n')
    def test_maskSeq(self):
        print('TEST CUT>')
        result = maskSeq(self.mask_primer, mode='cut', barcode=True)
        print(' ID> %s' % result.id)
        print('SEQ> %s\n' % result.seq)
        self.assertEqual(str(self.mask_seq_cut.seq), str(result.seq))
        self.assertEqual(self.mask_seq_cut.id, result.id)

        print('TEST MASK>')
        result = maskSeq(self.mask_primer, mode='mask', barcode=True)
        print(' ID> %s' % result.id)
        print('SEQ> %s\n' % result.seq)
        self.assertEqual(str(self.mask_seq_mask.seq), str(result.seq))
        self.assertEqual(self.mask_seq_mask.id, result.id)

        print('TEST TRIM>')
        result = maskSeq(self.mask_primer, mode='trim', barcode=True)
        print(' ID> %s' % result.id)
        print('SEQ> %s\n' % result.seq)
        self.assertEqual(str(self.mask_seq_trim.seq), str(result.seq))
        self.assertEqual(self.mask_seq_trim.id, result.id)

        print('TEST TAG>')
        result = maskSeq(self.mask_primer, mode='tag', barcode=True)
        print(' ID> %s' % result.id)
        print('SEQ> %s\n' % result.seq)
        self.assertEqual(str(self.mask_seq_tag.seq), str(result.seq))
        self.assertEqual(self.mask_seq_tag.id, result.id)

        print('TEST CUT BARCODLEN>')
        result = maskSeq(self.mask_primer, mode='cut', barcode=True, barcode_length=2)
        print(' ID> %s' % result.id)
        print('SEQ> %s\n' % result.seq)
        self.assertEqual(str(self.mask_seq_cut_barcodelen.seq), str(result.seq))
        self.assertEqual(self.mask_seq_cut_barcodelen.id, result.id)

if __name__ == '__main__':
    unittest.main()


"""
Unit tests for Annotation module
"""
# Info
__author__ = 'Jason Anthony Vander Heiden'

# Imports
import os
import sys
import time
import unittest
from collections import OrderedDict

# Presto imports
from presto.Annotation import collapseAnnotation, mergeAnnotation, renameAnnotation, getCoordKey

# Paths
test_path = os.path.dirname(os.path.realpath(__file__))


class TestAnnotation(unittest.TestCase):
    """
    Notes:
      illumina  @MISEQ:132:000000000-A2F3U:1:1101:14340:1555 2:N:0:ATCACG
                @HWI-EAS209_0006_FC706VJ:5:58:5894:21141#ATCACG/1
                @MS6_33112:1:1101:18371:1066/1
      sra       @SRR001666.1 071112_SLXA-EAS1_s_7:5:1:817:345 length=36
                @SRR001666.1.2 1 length=250
      454       @000034_0199_0169 length=437 uaccno=GNDG01201ARRCR
      pesto     @AATCGGATTTGC|COUNT=2|PRIMER=IGHJ_RT|PRFREQ=1.0
    """
    def setUp(self):
        print('-> %s()' % self._testMethodName)
        # Annotation dictionaries
        self.ann_dict_1 = OrderedDict([('ID', 'SEQ1'),
                                       ('TEST1', 'A,B'),
                                       ('TEST2', [1, 2]),
                                       ('TEST3', ['First','Second'])])
        self.ann_dict_2 = OrderedDict([('ID', 'SEQ2'),
                                       ('TEST1', 'C,C'),
                                       ('TEST2', 3)])

        # SRA, ENA
        self.desc_sra = ['SRR001666.1 071112_SLXA-EAS1_s_7:5:1:817:345 length=36',
                         'SRR735691.1 GXGJ56Z01AE06X length=222',
                         'SRR1383326.1 1 length=250',
                         'SRR1383326.1.2 1 length=250',
                         'ERR346596.6 BS-DSFCONTROL04:4:000000000-A3F0Y:1:1101:13220:1649/1']
        self.id_sra = ['SRR001666.1',
                       'SRR735691.1',
                       'SRR1383326.1',
                       'SRR1383326.1',
                       'ERR346596.6']

        # 454
        self.desc_454 = ['000034_0199_0169 length=437 uaccno=GNDG01201ARRCR',
                         'GXGJ56Z01AE06X length=222']
        self.id_454 = ['000034_0199_0169',
                       'GXGJ56Z01AE06X']

        # Illumina, Solexa
        self.desc_illumina = ['MISEQ:132:000000000-A2F3U:1:1101:14340:1555 1:N:0:ATCACG',
                              'HWI-EAS209_0006_FC706VJ:5:58:5894:21141#ATCACG/1',
                              'MS6_33112:1:1101:18371:1066/1']
        self.id_illumina = ['MISEQ:132:000000000-A2F3U:1:1101:14340:1555',
                            'HWI-EAS209_0006_FC706VJ:5:58:5894:21141',
                            'MS6_33112:1:1101:18371:1066']

        # Start clock
        self.start = time.time()

    def tearDown(self):
        # End clock
        t = time.time() - self.start
        print('<- %s() %.3f' % (self._testMethodName, t))

    #@unittest.skip('-> mergeAnnotation() skipped\n')
    def test_mergeAnnotation(self):
        result = mergeAnnotation(self.ann_dict_1, self.ann_dict_2)
        print(result)
        self.assertEqual('SEQ1', result['ID'])
        self.assertEqual('A,B,C,C', result['TEST1'])
        self.assertEqual('1,2,3', result['TEST2'])

        result = mergeAnnotation(self.ann_dict_1, self.ann_dict_2, prepend=True)
        print(result)
        self.assertEqual('SEQ1', result['ID'])
        self.assertEqual('C,C,A,B', result['TEST1'])
        self.assertEqual('3,1,2', result['TEST2'])

    #@unittest.skip('-> renameAnnotation() skipped\n')
    def test_renameAnnotation(self):
        result = renameAnnotation(self.ann_dict_1, 'TEST1', 'RENAMED1')
        print(result)
        self.assertEqual('A,B', result['RENAMED1'])
        self.assertEqual(None, result.get('TEST1', None))

        result = renameAnnotation(self.ann_dict_1, 'TEST1', 'RENAMED2')
        print(result)
        self.assertEqual('A,B', result['RENAMED2'])
        self.assertEqual(None, result.get('TEST1', None))

        result = renameAnnotation(self.ann_dict_1, 'TEST2', 'TEST1')
        print(result)
        self.assertEqual('A,B,1,2', result['TEST1'])
        self.assertEqual(None, result.get('TEST2', None))

        result = renameAnnotation(self.ann_dict_1, 'TEST2', 'TEST1')
        result = renameAnnotation(result, 'TEST3', 'TEST1')
        print(result)
        self.assertEqual('A,B,1,2,First,Second', result['TEST1'])
        self.assertEqual(None, result.get('TEST2', None))
        self.assertEqual(None, result.get('TEST3', None))

    #@unittest.skip('-> collapseAnnotation() skipped\n')
    def test_collapseAnnotation(self):
        result = collapseAnnotation(self.ann_dict_1, 'first')
        print(result)
        self.assertEqual('SEQ1', result['ID'])
        self.assertEqual('A', result['TEST1'])
        self.assertEqual(1, result['TEST2'])

        result = collapseAnnotation(self.ann_dict_1, 'last')
        print(result)
        self.assertEqual('SEQ1', result['ID'])
        self.assertEqual('B', result['TEST1'])
        self.assertEqual(2, result['TEST2'])

        result = collapseAnnotation(self.ann_dict_1, 'set')
        self.assertEqual('SEQ1', result['ID'])
        self.assertEqual(['A' ,'B'], result['TEST1'])
        self.assertEqual([1, 2], result['TEST2'])
        print(result)

        result = collapseAnnotation(self.ann_dict_1, 'cat')
        self.assertEqual('SEQ1', result['ID'])
        self.assertEqual('AB', result['TEST1'])
        self.assertEqual('12', result['TEST2'])
        print(result)

        # TODO:  may be better for min/max/sum to return float/int
        result = collapseAnnotation(self.ann_dict_1, 'min', fields='TEST2')
        self.assertEqual('1', result['TEST2'])
        print(result)

        result = collapseAnnotation(self.ann_dict_1, 'max', fields='TEST2')
        self.assertEqual('2', result['TEST2'])
        print(result)

        result = collapseAnnotation(self.ann_dict_1, 'sum', fields='TEST2')
        self.assertEqual('3', result['TEST2'])
        print(result)

    #@unittest.skip('-> getCoordKey() skipped\n')
    def test_getCoordKey(self):
        result = [getCoordKey(x, coord_type='illumina') for x in self.desc_illumina]
        self.assertListEqual(result, self.id_illumina)
        print(result)

        result = [getCoordKey(x, coord_type='sra') for x in self.desc_sra]
        self.assertListEqual(result, self.id_sra)
        print(result)

        result = [getCoordKey(x, coord_type='454') for x in self.desc_454]
        self.assertListEqual(result, self.id_454)
        print(result)



if __name__ == '__main__':
    unittest.main()


#!/usr/bin/env python3
"""
Converts sequence headers to the pRESTO format
"""
# Info
__author__ = 'Jason Anthony Vander Heiden'
from presto import __version__, __date__

# Imports
import os
from argparse import ArgumentParser
from collections import OrderedDict
from textwrap import dedent
from time import time
from Bio import SeqIO

# Presto imports
from presto.Defaults import default_out_args
from presto.Annotation import flattenAnnotation, convertGenericHeader, convert454Header, \
                              convertGenbankHeader, convertIlluminaHeader, convertIMGTHeader, \
                              convertSRAHeader, convertMIGECHeader
from presto.Commandline import CommonHelpFormatter, checkArgs, getCommonArgParser, parseCommonArgs
from presto.IO import getFileType, readSeqFile, countSeqFile, getOutputHandle, \
                      printLog, printProgress


def convertHeaders(seq_file, convert_func, convert_args={}, out_file=None, out_args=default_out_args):
    """
    Converts sequence headers to the pRESTO format

    Arguments:
      seq_file : the sequence file name.
      convert_func : the function used to convert sequence headers.
      convert_args : a dictionary of arguments to pass to convert_func.
      out_file : output file name. Automatically generated from the input file if None.
      out_args : common output argument dictionary from parseCommonArgs.

    Returns:
      str: the output sequence file name.
    """
    # Define subcommand label dictionary
    cmd_dict = {convertGenericHeader: 'generic',
                convert454Header: '454',
                convertGenbankHeader: 'genbank',
                convertIlluminaHeader: 'illumina',
                convertIMGTHeader: 'imgt',
                convertMIGECHeader: 'migec',
                convertSRAHeader: 'sra'}

    log = OrderedDict()
    log['START'] = 'ConvertHeaders'
    log['COMMAND'] = cmd_dict[convert_func]
    log['FILE'] = os.path.basename(seq_file)
    printLog(log)

    # Open input file
    in_type = getFileType(seq_file)
    seq_iter = readSeqFile(seq_file)
    if out_args['out_type'] is None:  out_args['out_type'] = in_type

    # Wrapper for opening handles and writers
    def _open(x, out_file=out_file):
        if out_file is not None and x == 'pass':
            handle = open(out_file, 'w')
        else:
            handle = getOutputHandle(seq_file,
                                     'convert-%s' % x,
                                     out_dir=out_args['out_dir'],
                                     out_name=out_args['out_name'],
                                     out_type=out_args['out_type'])
        return handle

    # Count records
    result_count = countSeqFile(seq_file)

    # Set additional conversion arguments
    if convert_func in [convertGenericHeader, convertGenbankHeader]:
        convert_args.update({'delimiter': out_args['delimiter']})

    # Intialize file handles
    pass_handle, fail_handle = None, None

    # Iterate over sequences
    start_time = time()
    seq_count = pass_count = fail_count = 0
    for seq in seq_iter:
        # Print progress for previous iteration and update count
        printProgress(seq_count, result_count, 0.05, start_time=start_time)
        seq_count += 1

        # Convert header
        header = convert_func(seq.description, **convert_args)

        if header is not None:
            # Write successfully converted sequences
            pass_count += 1
            seq.id = seq.name = flattenAnnotation(header, out_args['delimiter'])
            seq.description = ''
            try:
                SeqIO.write(seq, pass_handle, out_args['out_type'])
            except AttributeError:
                # Open output file
                pass_handle = _open('pass')
                SeqIO.write(seq, pass_handle, out_args['out_type'])
        else:
            fail_count += 1
            if out_args['failed']:
                # Write unconverted sequences
                try:
                    SeqIO.write(seq, fail_handle, out_args['out_type'])
                except AttributeError:
                    # Open output file
                    pass_handle = _open('fail')
                    SeqIO.write(seq, fail_handle, out_args['out_type'])

    # Print counts
    printProgress(seq_count, result_count, 0.05, start_time=start_time)
    log = OrderedDict()
    log['OUTPUT'] = os.path.basename(pass_handle.name) if pass_handle is not None else None
    log['SEQUENCES'] = seq_count
    log['PASS'] = pass_count
    log['FAIL'] = fail_count
    log['END'] = 'ConvertHeaders'
    printLog(log)

    # Close file handles
    if fail_handle is not None:  pass_handle.close()
    if fail_handle is not None:  fail_handle.close()

    return pass_handle.name


def getArgParser():
    """
    Defines the ArgumentParser

    Returns:
      argparse.ArgumentParser: argument parser object.
    """
    # Define output file names and header fields
    fields = dedent(
             '''
             output files:
                 convert-pass
                     reads passing header conversion.
                 convert-fail
                     raw reads failing header conversion.

             output annotation fields:
                 <format defined>
                     the annotation fields added are specific to the header format of the
                     input file.
             ''')

    # Define ArgumentParser
    parser = ArgumentParser(description=__doc__, epilog=fields,
                            formatter_class=CommonHelpFormatter, add_help=False)
    group_help = parser.add_argument_group('help')
    group_help.add_argument('--version', action='version',
                            version='%(prog)s:' + ' %s %s' %(__version__, __date__))
    group_help.add_argument('-h', '--help', action='help', help='show this help message and exit')
    subparsers = parser.add_subparsers(title='subcommands', metavar='',
                                       help='Conversion method')
    # TODO:  This is a temporary fix for Python issue 9253
    subparsers.required = True

    # Parent parser defining universal argument
    parent_parser = getCommonArgParser(log=False)

    # Subparser for generic header conversion
    parser_generic = subparsers.add_parser('generic', parents=[parent_parser],
                                           formatter_class=CommonHelpFormatter, add_help=False,
                                           help='''Converts sequence headers without a known
                                            annotation system.''',
                                           description='''Converts sequence headers without a known
                                            annotation system.''')
    parser_generic.set_defaults(convert_func=convertGenericHeader)

    # Subparser for conversion of 454 headers
    parser_454 = subparsers.add_parser('454', parents=[parent_parser],
                                       formatter_class=CommonHelpFormatter, add_help=False,
                                       help='''Converts Roche 454 sequence headers.''',
                                       description='''Converts Roche 454 sequence headers.''')
    parser_454.set_defaults(convert_func=convert454Header)

    # Subparser for conversion of GenBank and RefSeq headers
    parser_genbank = subparsers.add_parser('genbank', parents=[parent_parser],
                                           formatter_class=CommonHelpFormatter, add_help=False,
                                           help='''Converts NCBI GenBank and RefSeq
                                                sequence headers.''',
                                           description='''Converts NCBI GenBank and RefSeq
                                                sequence headers.''')
    parser_genbank.set_defaults(convert_func=convertGenbankHeader)

    # Subparser for conversion of Illumina headers
    parser_illumina = subparsers.add_parser('illumina', parents=[parent_parser],
                                            formatter_class=CommonHelpFormatter, add_help=False,
                                            help='''Converts Illumina sequence headers.''',
                                            description='''Converts Illumina sequence headers.''')
    parser_illumina.set_defaults(convert_func=convertIlluminaHeader)

    # Subparser for conversion of IMGT germline headers
    parser_imgt = subparsers.add_parser('imgt', parents=[parent_parser],
                                        formatter_class=CommonHelpFormatter, add_help=False,
                                        help='''Converts sequence headers output by
                                             IMGT/GENE-DB.''',
                                        description='''Converts sequence headers output by
                                             IMGT/GENE-DB.''')
    group_imgt = parser_imgt.add_argument_group('conversion arguments')
    group_imgt.add_argument('--simple', action='store_true', dest='simple',
                            help='''If specified, only the allele name, and no other
                                 annotations, will appear in the converted sequence
                                 header.''')
    parser_imgt.set_defaults(convert_func=convertIMGTHeader)

    # Subparser for conversion of MIGEC headers
    parser_migec = subparsers.add_parser('migec', parents=[parent_parser],
                                         formatter_class=CommonHelpFormatter, add_help=False,
                                         help='''Converts headers for consensus sequence generated
                                              by the MIGEC tool.''',
                                         description='''Converts headers for consensus sequence generated
                                              by the MIGEC tool.''')
    parser_migec.set_defaults(convert_func=convertMIGECHeader)

    # Subparser for conversion of SRA headers
    parser_sra = subparsers.add_parser('sra', parents=[parent_parser], add_help=False,
                                       formatter_class=CommonHelpFormatter,
                                       help='''Converts NCBI SRA or EMBL-EBI ENA sequence headers.''',
                                       description='''Converts NCBI SRA or EMBL-EBI ENA sequence headers.''')
    parser_sra.set_defaults(convert_func=convertSRAHeader)

    return parser


if __name__ == '__main__':
    """
    Parses command line arguments and calls main function
    """
    # Parse arguments
    parser = getArgParser()
    checkArgs(parser)
    args = parser.parse_args()
    args_dict = parseCommonArgs(args)

    # Create convert_args
    convert_keys = ['simple']
    args_dict['convert_args'] = dict((k, args_dict[k]) for k in args_dict \
                                     if k in convert_keys)
    for k in args_dict['convert_args']:  del args_dict[k]

    # Calls header conversion function
    del args_dict['seq_files']
    if 'out_files' in args_dict:  del args_dict['out_files']
    for i, f in enumerate(args.__dict__['seq_files']):
        args_dict['seq_file'] = f
        args_dict['out_file'] = args.__dict__['out_files'][i] \
            if args.__dict__['out_files'] else None
        convertHeaders(**args_dict)



"""
Unit tests for ConvertHeaders
"""
# Info
__author__ = 'Jason Anthony Vander Heiden'

# Imports
import os
import sys
import time
import unittest

# Paths
test_path = os.path.dirname(os.path.realpath(__file__))

# Import script
sys.path.append(os.path.join(test_path, os.pardir, 'bin'))
import ConvertHeaders


class TestConvertHeaders(unittest.TestCase):
    def setUp(self):
        print('-> %s()' % self._testMethodName)

        # IMGT
        self.desc_imgt = ['M99641|IGHV1-18*01|Homo sapiens|F|V-REGION|188..483|296 nt|1| | | | |296+24=320| | |',
						  'Z29978|IGHV1-69*07|Homo sapiens|F|V-REGION|1..233|233 nt|1| | | | |233+58=291|partial in 5\' and in 3\' | |',
						  'AB019439|IGHV(III)-22-2*01|Homo sapiens|P|L-PART1+V-EXON|168065..168119+168222..168262|96 nt|1| | | | |96+0=96| | |',
						  'AE000659|TRAV11*01|Homo sapiens|(P)|V-REGION|85121..85397|276 nt|1| | | | |276+42=318| | |']

        # GenBank
        self.desc_genbank = ['gi|568336023|gb|CM000663.2| Homo sapiens chromosome 1, GRCh38 reference primary assembly',
                             'CM000663.2 Homo sapiens chromosome 1, GRCh38 reference primary assembly']

        # SRA, ENA
        self.desc_sra = ['SRR001666.1 071112_SLXA-EAS1_s_7:5:1:817:345 length=36',
                         'SRR735691.1 GXGJ56Z01AE06X length=222',
                         'SRR1383326.1 1 length=250',
                         'SRR1383326.1.2 1 length=250',
                         'ERR346596.6 BS-DSFCONTROL04:4:000000000-A3F0Y:1:1101:13220:1649/1']

        # 454
        self.desc_454 = ['000034_0199_0169 length=437 uaccno=GNDG01201ARRCR',
                         'GXGJ56Z01AE06X length=222']

        # Illumina, Solexa
        self.desc_illumina = ['MISEQ:132:000000000-A2F3U:1:1101:14340:1555 1:N:0:ATCACG',
                              'HWI-EAS209_0006_FC706VJ:5:58:5894:21141#ATCACG/1',
                              'MS6_33112:1:1101:18371:1066/1']

        # Start clock
        self.start = time.time()

    def tearDown(self):
        # End clock
        t = time.time() - self.start
        print('<- %s() %.3f' % (self._testMethodName, t))

    #@unittest.skip('-> convertGenericHeader() skipped\n')
    def test_convertGenericHeader(self):
        results = [ConvertHeaders.convertGenericHeader(x) for x in self.desc_genbank]
        for x in results:
            print('%s' % x)

        self.assertTrue(all([x is not None for x in results]))

        results = [ConvertHeaders.convertGenericHeader(x) for x in self.desc_imgt]
        for x in results:
            print('%s' % x)

        self.assertTrue(all([x is not None for x in results]))

    #@unittest.skip('-> convertIMGTHeader() skipped\n')
    def test_convertIMGTHeader(self):
        results = [ConvertHeaders.convertIMGTHeader(x) for x in self.desc_imgt]
        for x in results:
            print('%s' % x)

        self.assertTrue(all([x is not None for x in results]))

        results = [ConvertHeaders.convertIMGTHeader(x, simple=True) for x in self.desc_imgt]
        for x in results:
            print('%s' % x)

        self.assertTrue(all([x is not None for x in results]))

    #@unittest.skip('-> convertIlluminaHeader() skipped\n')
    def test_convertIlluminaHeader(self):
        results = [ConvertHeaders.convertIlluminaHeader(x) for x in self.desc_illumina]
        for x in results:
            print('%s' % x)

        self.assertTrue(all([x is not None for x in results]))

    #@unittest.skip('-> convertGenbankHeader() skipped\n')
    def test_convertGenbankHeader(self):
        results = [ConvertHeaders.convertGenbankHeader(x) for x in self.desc_genbank]
        for x in results:
            print('%s' % x)

        self.assertTrue(all([x is not None for x in results]))

    #@unittest.skip('-> convertSRAHeader() skipped\n')
    def test_convertSRAHeader(self):
        results = [ConvertHeaders.convertSRAHeader(x) for x in self.desc_sra]
        for x in results:
            print('%s' % x)

        self.assertTrue(all([x is not None for x in results]))

    #@unittest.skip('-> convert454Header() skipped\n')
    def test_convert454Header(self):
        results = [ConvertHeaders.convert454Header(x) for x in self.desc_454]
        for x in results:
            print('%s' % x)

        self.assertTrue(all([x is not None for x in results]))


if __name__ == '__main__':
    unittest.main()


"""
Default parameters
"""
# Info
__author__ = 'Jason Anthony Vander Heiden, Namita Gupta'

# Annotation parameters
default_delimiter = ('|', '=', ',')
default_separator = default_delimiter[2]

# Commandline argument defaults
choices_coord = ['illumina', 'solexa', 'sra', '454', 'presto']
default_coord = 'presto'
default_out_args = {'log_file':None,
                    'delimiter':default_delimiter,
                    'separator':default_separator,
                    'out_dir':None,
                    'out_name':None,
                    'out_type':None,
                    'failed':True}

# Fields
default_barcode_field = 'BARCODE'
default_primer_field = 'PRIMER'
default_cluster_field = 'CLUSTER'

# External applications
default_muscle_exec = 'muscle'
default_usearch_exec = 'usearch'
default_vsearch_exec = 'vsearch'
default_cdhit_exec = 'cd-hit-est'
default_blastn_exec = 'blastn'
default_blastdb_exec = 'makeblastdb'

# Sequence sets
default_gap_chars = set(['-', '.'])
default_mask_chars = set(['n', 'N'])
default_missing_chars = set(['-', '.', 'n', 'N'])
default_missing_residues = set(['.', '-', 'x', 'N'])

# Consensus defaults
default_consensus_min_freq = 0.6
default_consensus_min_qual = 0

# Primer defaults
default_primer_gap_penalty = (1, 1)
default_primer_max_error = 0.2
default_primer_max_len = 50
default_primer_start = 0

# Filter defaults
default_filter_min_qual = 20
default_filter_max_missing = 10
default_filter_max_repeat = 15
default_filter_min_len = 250
default_filter_window = 10

# Assembly defaults
default_assembly_alpha = 1e-5
default_assembly_max_error = 0.3
default_assembly_min_ident = 0.5
default_assembly_min_len = 8
default_assembly_max_len = 1000
default_assembly_gap = 0
default_assembly_evalue = 1e-5
default_assembly_max_hits = 100
default_consensus_min_count = 1


#!/usr/bin/env python3
"""
Removes duplicate sequences from FASTA/FASTQ files
"""
# Info
__author__ = 'Jason Anthony Vander Heiden'
from presto import __version__, __date__

# Imports
import os
import re
from argparse import ArgumentParser
from collections import OrderedDict
from itertools import chain
from textwrap import dedent
from time import time
from Bio import SeqIO

# Presto imports
from presto.Defaults import default_delimiter, default_out_args
from presto.Commandline import CommonHelpFormatter, checkArgs, getCommonArgParser, parseCommonArgs
from presto.Annotation import parseAnnotation, flattenAnnotation, mergeAnnotation, \
                              collapseAnnotation
from presto.Sequence import checkSeqEqual
from presto.IO import getFileType, readSeqFile, getOutputHandle, printLog, printProgress

# Default parameters
default_max_missing = 0

class DuplicateSet:
    """
    A class defining unique sequence sets

    Attributes:
      seq : SeqRecord.
      missing : missing character count of SeqRecord.
      annotation: copy field annotation dictionary.
      keys : list of sequence identifier.
      count : duplicate count.
    """
    # Instantiation
    def __init__(self, seq, key, missing, annotations):
        self.seq = seq
        self.missing = missing
        self.annotations = annotations
        self.keys = [key]
        self.count = 1

    # Set length evaluation to number of duplicates
    def __len__(self):
        return len(self.keys)


def findUID(uid, search_dict, score=False):
    """
    Checks if a unique identifier is already present in a unique dictionary

    Arguments:
      uid : the unique identifier key to check.
      search_dict : a dictionary to search for key matches in.
      score : if True score sequence element of the uid against each sequence in search_list.

    Returns:
      str: uid of match if found; None otherwise.
    """
    match = None
    # Check for exact matches
    if not score:
        match = uid if uid in search_dict else None
    # Check for ambiguous matches
    else:
        for key in search_dict:
            if uid[1:] == key[1:] and checkSeqEqual(uid[0], key[0]):
                match = key
                break

    # Return search boolean
    return match


def findUniqueSeq(uniq_dicts, search_keys, seq_dict, max_missing=default_max_missing, 
                  uniq_fields=None, copy_fields=None, max_field=None, min_field=None,
                  inner=False, delimiter=default_delimiter):
    """
    Finds unique sequences

    Arguments:
      uniq_dicts: a dictionary of length segregated dictionaries of unique sequences generated by findUniqueSeq().
      search_keys : a list containing the subset of dictionary keys to be checked.
      seq_dict : a SeqRecords dictionary generated by SeqIO.index().
      max_missing : the number of missing characters to allow in a unique sequences.
      uniq_fields : a list of annotations that define a sequence as unique if they differ.
      copy_fields : a list of annotations to copy into unique sequence annotations.
      max_field : a numeric field whose maximum value determines the retained sequence.
      min_field : a numeric field whose minimum value determines the retained sequence.
      inner : if True exclude consecutive outer ambiguous characters from iterations and matching.
      delimiter : description field delimiter.

    Returns:
      tuple: (uniq_dict, search_keys, dup_keys) modified from passed values.

    This function was modified to improve performance by segregating uniq sequences into separate dicts.  Each dict holds 
    all unique sequences of a specific length.  
    """
    # Define local variables
    ambig_re = re.compile(r'[\.\-N]')
    score = (max_missing > 0)
    dup_keys = []
    to_remove = []

    start_time = time()
    result_count = len(search_keys)
    # Iterate over search keys and update uniq_dict and dup_keys
    for idx, key in enumerate(search_keys):
        # Print progress of previous iteration
        printProgress(idx, result_count, 0.05, start_time=start_time, task='%i missing' % max_missing)

        # Define sequence to process
        seq = seq_dict[key]
        seq_str = str(seq.seq)
        if inner:  seq_str = seq_str.strip('.-N')

        # Skip processing of ambiguous sequences over max_missing threshold
        ambig_count = len(ambig_re.findall(seq_str))
        if ambig_count > max_missing:  continue

        # Parse annotation and define unique identifiers (uid)
        if uniq_fields is not None:
            ann = parseAnnotation(seq_dict[key].description, uniq_fields, delimiter=delimiter)
            uid = tuple(chain([seq_str], list(ann.values())))
        else:
            uid = (seq_str, None)

        # Parse annotation and define copied identifiers (cid)
        if copy_fields is not None:
            ann = parseAnnotation(seq.description, copy_fields, delimiter=delimiter)
            cid = {k:[ann.get(k)] for k in copy_fields}
        else:
            cid = {}

        ##  use or create length specific uniq_dict
        seq_len=len(uid[0])
        uniq_dict=uniq_dicts.setdefault(seq_len, {})
        
        # Store new unique sequences and process duplicates
        match = findUID(uid, uniq_dict, score)
        if match is None:
            uniq_dict[uid] = DuplicateSet(seq, key=key, missing=ambig_count, annotations=cid)
        else:
            # Updated sequence, count, ambiguous character count, and count sets
            dup_key = key
            uniq_dict[match].count += 1
            uniq_dict[match].keys.append(key)
            for k, v in cid.items():
                uniq_dict[match].annotations[k].extend(v)
            # Check whether to replace previous unique sequence with current sequence
            if ambig_count <= uniq_dict[match].missing:
                swap = False
                seq_last = uniq_dict[match].seq
                if max_field is not None:
                    swap = float(parseAnnotation(seq.description, delimiter=delimiter)[max_field]) > \
                           float(parseAnnotation(seq_last.description, delimiter=delimiter)[max_field])
                elif min_field is not None:
                    swap = float(parseAnnotation(seq.description, delimiter=delimiter)[min_field]) > \
                           float(parseAnnotation(seq_last.description, delimiter=delimiter)[min_field])
                # TODO:  quality evaluation is a bottleneck
                else:
                    if hasattr(seq, 'letter_annotations') and 'phred_quality' in seq.letter_annotations:
                        q_this = float(sum(seq.letter_annotations['phred_quality'])) / len(seq)
                        q_last = float(sum(seq_last.letter_annotations['phred_quality'])) / len(seq_last)
                        swap = q_this > q_last
                # Replace old sequence if criteria passed
                if swap:
                    dup_key = seq_last.id
                    uniq_dict[match].seq = seq
                    uniq_dict[match].missing = ambig_count

            # Update duplicate list
            dup_keys.append(dup_key)

        # Mark seq for removal from later steps
        to_remove.append(idx)

    # Remove matched sequences from search_keys
    for j in reversed(to_remove):  del search_keys[j]

    # Update progress
    printProgress(result_count, result_count, 0.05, start_time=start_time, task='%i missing' % max_missing)

    return (uniq_dicts, search_keys, dup_keys)  


def merge_dicts(dicts):
    print(f"merging {len(dicts)} dicts")
    merged_dict = {k:v for l, d in dicts.items() for k, v in d.items()}
    return merged_dict

def collapseSeq(seq_file, max_missing=default_max_missing, uniq_fields=None,
                copy_fields=None, copy_actions=None, max_field=None, min_field=None,
                inner=False, keep_missing=False, out_file=None, out_args=default_out_args):
    """
    Removes duplicate sequences from a file

    Arguments:
      seq_file : filename of the sequence file to sample from.
      max_missing : number of ambiguous characters to allow in a unique sequence.
      uniq_fields : a list of annotations that define a sequence as unique if they differ.
      copy_fields : a list of annotations to copy into unique sequence annotations.
      copy_actions : the list of collapseAnnotation actions to take on copy_fields.
      max_field : a numeric field whose maximum value determines the retained sequence.
      min_field : a numeric field whose minimum value determines the retained sequence.
      inner : if True exclude consecutive outer ambiguous characters from iterations and matching.
      keep_missing : if True retain sequences with more ambiguous characters than max_missing as unique.
      out_file : output file name. Automatically generated from the input file if None.
      out_args : common output argument dictionary from parseCommonArgs.

    Returns:
      str: the collapsed output file name.
    """
    log = OrderedDict()
    log['START'] = 'CollapseSeq'
    log['FILE'] = os.path.basename(seq_file)
    log['MAX_MISSING'] = max_missing
    log['UNIQ_FIELDS'] = ','.join([str(x) for x in uniq_fields]) \
                         if uniq_fields is not None else None
    log['COPY_FIELDS'] = ','.join([str(x) for x in copy_fields]) \
                         if copy_fields is not None else None
    log['COPY_ACTIONS'] = ','.join([str(x) for x in copy_actions]) \
                          if copy_actions is not None else None
    log['MAX_FIELD'] = max_field
    log['MIN_FIELD'] = min_field
    log['INNER'] = inner
    log['KEEP_MISSING'] = keep_missing
    printLog(log)

    # Read input file
    in_type = getFileType(seq_file)
    seq_dict = SeqIO.to_dict(readSeqFile(seq_file, index=False))
    if out_args['out_type'] is None:  out_args['out_type'] = in_type

    # Count total sequences
    rec_count = len(seq_dict)

    # Open unique record output handle
    if out_file is not None:
        pass_handle = open(out_file, 'w')
    else:
        pass_handle = getOutputHandle(seq_file,
                                      'collapse-unique',
                                      out_dir=out_args['out_dir'],
                                      out_name=out_args['out_name'],
                                      out_type=out_args['out_type'])
    # Define log handle
    if out_args['log_file'] is None:
        log_handle = None
    else:
        log_handle = open(out_args['log_file'], 'w')

    # Find sequences with duplicates
    uniq_dicts = {} 
    # Added list typing for compatibility issue with Python 2.7.5 on OS X
    # TypeError: object of type 'dictionary-keyiterator' has no len()
    search_keys = list(seq_dict.keys())
    dup_keys = []
    for n in range(0, max_missing + 1):
        # Find unique sequences
        uniq_dicts, search_keys, dup_list = findUniqueSeq(uniq_dicts, search_keys, seq_dict, n, 
                                                         uniq_fields, copy_fields,
                                                         max_field, min_field, inner,
                                                         out_args['delimiter'])

        # Update list of duplicates
        dup_keys.extend(dup_list)

        # Break if no keys to search remain
        if len(search_keys) == 0:  break

    ## merge all uniq_dicts into one dict here.
    uniq_dict=merge_dicts(uniq_dicts)

    # Write unique sequences
    for val in uniq_dict.values():
        # Define output sequence
        out_seq = val.seq
        out_ann = parseAnnotation(out_seq.description, delimiter=out_args['delimiter'])
        out_app = OrderedDict()
        if copy_fields  is not None and copy_actions is not None:
            for f, a in zip(copy_fields, copy_actions):
                x = collapseAnnotation(val.annotations, a, f, delimiter=out_args['delimiter'])
                out_app[f] = x[f]
                out_ann.pop(f, None)
        out_app['DUPCOUNT'] = val.count
        out_ann = mergeAnnotation(out_ann, out_app, delimiter=out_args['delimiter'])
        out_seq.id = out_seq.name = flattenAnnotation(out_ann, delimiter=out_args['delimiter'])
        out_seq.description = ''
        # Write unique sequence
        SeqIO.write(out_seq, pass_handle, out_args['out_type'])

        # Update log
        log = OrderedDict()
        log['HEADER'] = out_seq.id
        log['DUPCOUNT'] = val.count
        for i, k in enumerate(val.keys, start=1):
            log['ID%i' % i] = k
        for i, k in enumerate(val.keys, start=1):
            log['SEQ%i' % i] = str(seq_dict[k].seq)
        printLog(log, handle=log_handle)

    # Write sequence with high missing character counts
    if keep_missing:
        for k in search_keys:
            out_seq = seq_dict[k]
            out_ann = parseAnnotation(out_seq.description, delimiter=out_args['delimiter'])
            out_ann = mergeAnnotation(out_ann, {'DUPCOUNT':1}, delimiter=out_args['delimiter'])
            out_seq.id = out_seq.name = flattenAnnotation(out_ann, delimiter=out_args['delimiter'])
            out_seq.description = ''
            SeqIO.write(out_seq, pass_handle, out_args['out_type'])

    # Write sequence with high missing character counts
    if out_args['failed'] and not keep_missing:
        with getOutputHandle(seq_file, 'collapse-undetermined', out_dir=out_args['out_dir'],
                             out_name=out_args['out_name'], out_type=out_args['out_type']) \
                as missing_handle:
            for k in search_keys:
                SeqIO.write(seq_dict[k], missing_handle, out_args['out_type'])

    if out_args['failed']:
        # Write duplicate sequences
        with getOutputHandle(seq_file, 'collapse-duplicate', out_dir=out_args['out_dir'],
                             out_name=out_args['out_name'], out_type=out_args['out_type']) \
                as dup_handle:
            for k in dup_keys:
                SeqIO.write(seq_dict[k], dup_handle, out_args['out_type'])

    # Print log
    log = OrderedDict()
    log['OUTPUT'] = os.path.basename(pass_handle.name)
    log['SEQUENCES'] = rec_count
    log['UNIQUE'] = len(uniq_dict)
    log['DUPLICATE'] = len(dup_keys)
    log['UNDETERMINED'] = len(search_keys)
    log['END'] = 'CollapseSeq'
    printLog(log)

    # Close file handles
    pass_file = pass_handle.name
    if pass_handle is not None:  pass_handle.close()
    if log_handle is not None:  log_handle.close()

    return pass_file


def getArgParser():
    """
    Defines the ArgumentParser

    Returns:
      argparse.ArgumentParser: argument parser object.
    """
    # Define output file names and header fields
    fields = dedent(
             '''
             output files:
                 collapse-unique
                     unique sequences. Contains one representative from each set of
                     duplicate sequences. The retained representative is determined by
                     user defined criteria.
                 collapse-duplicate
                     raw reads which are duplicates of the sequences retained in the
                     collapse-unique file.
                 collapse-undetermined
                     raw reads which were excluded from consideration due to having too
                     many N characters in the sequence.

             output annotation fields:
                 DUPCOUNT
                     total number of sequences within the set of duplicates for each
                     retained unique sequence. Meaning, the copy number of each unique
                     sequence within the data file.
                 <user defined>
                     annotation fields specified by the --cf parameter.
             ''')

    # TODO: add exact mode which does hash table search only
    # TODO: write better algorithm for ambiguous character mode
    # Define ArgumentParser
    parser = ArgumentParser(description=__doc__, epilog=fields,
                            parents=[getCommonArgParser()],
                            formatter_class=CommonHelpFormatter, add_help=False)

    # Collapse arguments
    group_dedup = parser.add_argument_group('collapse arguments')
    group_dedup.add_argument('-n', action='store', dest='max_missing', type=int, default=default_max_missing,
                             help='''Maximum number of missing nucleotides to consider for collapsing
                                   sequences. A sequence will be considered undetermined if it contains too
                                   many missing nucleotides. Note, setting a value above 0 will consider
                                   ambiguous/missing nucleotides via a distance calculation, but is considerably
                                   more computationally expensive, especially on large data sets.''')
    group_dedup.add_argument('--uf', nargs='+', action='store', dest='uniq_fields', type=str, default=None,
                             help='''Specifies a set of annotation fields that must match for sequences
                                  to be considered duplicates.''')
    group_dedup.add_argument('--cf', nargs='+', action='store', dest='copy_fields', type=str, default=None,
                             help='''Specifies a set of annotation fields to copy into the unique
                                  sequence output.''')
    group_dedup.add_argument('--act', nargs='+', action='store', dest='copy_actions', default=None,
                             choices=['min', 'max', 'sum', 'set'],
                             help='''List of actions to take for each copy field which defines how
                                  each annotation will be combined into a single value. The actions
                                  "min", "max", "sum" perform the corresponding mathematical
                                  operation on numeric annotations. The action "set" collapses
                                  annotations into a comma delimited list of unique values.''')
    group_dedup.add_argument('--inner', action='store_true', dest='inner',
                             help='''If specified, exclude consecutive missing characters at either end of
                                  the sequence.''')
    group_dedup.add_argument('--keepmiss', action='store_true', dest='keep_missing',
                             help='''If specified, sequences with more missing characters than the
                                  threshold set by the -n parameter will be written to the unique
                                  sequence output file with a DUPCOUNT=1 annotation. If not specified,
                                  such sequences will be written to a separate file.''')

    # Mutually exclusive argument group
    group_field = group_dedup.add_mutually_exclusive_group()
    group_field.add_argument('--maxf', action='store', dest='max_field', type=str, default=None,
                             help='''Specify the field whose maximum value determines the retained sequence;
                                  mutually exclusive with --minf.''')
    group_field.add_argument('--minf', action='store', dest='min_field', type=str, default=None,
                             help='''Specify the field whose minimum value determines the retained sequence;
                                   mutually exclusive with --minf.''')

    return parser


if __name__ == '__main__':
    """
    Parses command line arguments and calls main function
    """
    # Parse arguments
    parser = getArgParser()
    checkArgs(parser)
    args = parser.parse_args()
    args_dict = parseCommonArgs(args)

    # Convert case of fields
    if 'uniq_fields' in args_dict and args_dict['uniq_fields']:
        args_dict['uniq_fields'] = list(map(str.upper, args_dict['uniq_fields']))
    if 'copy_fields' in args_dict and args_dict['copy_fields']:
        args_dict['copy_fields'] = list(map(str.upper, args_dict['copy_fields']))
    if 'copy_actions' in args_dict and args_dict['copy_actions']:
        args_dict['copy_actions'] = list(map(str.lower, args_dict['copy_actions']))
    if 'max_field' in args_dict and args_dict['max_field']:
        args_dict['max_field'] = args_dict['max_field'].upper()
    if 'min_field' in args_dict and args_dict['min_field']:
        args_dict['min_field'] = args_dict['min_field'].upper()

    # Check copy field and action arguments
    if bool(args_dict['copy_fields']) ^ bool(args_dict['copy_actions']) or \
       len((args_dict['copy_fields'] or '')) != len((args_dict['copy_actions'] or '')):
            parser.error('You must specify exactly one copy action (--act) per copy field (--cf)')

    # Call appropriate function for each sample file
    del args_dict['seq_files']
    if 'out_files' in args_dict:  del args_dict['out_files']
    for i, f in enumerate(args.__dict__['seq_files']):
        args_dict['seq_file'] = f
        args_dict['out_file'] = args.__dict__['out_files'][i] \
            if args.__dict__['out_files'] else None
        collapseSeq(**args_dict)



#!/usr/bin/env python3
"""
Cluster sequences by group
"""
# Info
__author__ = 'Christopher Bolen, Jason Anthony Vander Heiden, Ruoyi Jiang'
from presto import __version__, __date__

# Imports
import os
import shutil
import sys
from argparse import ArgumentParser
from collections import OrderedDict
from textwrap import dedent
from time import time
from Bio import SeqIO
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord

# Presto imports
from presto.Defaults import default_delimiter, default_barcode_field, \
                            default_cluster_field, default_out_args, \
                            default_usearch_exec, default_vsearch_exec, default_cdhit_exec
from presto.Commandline import CommonHelpFormatter, checkArgs, getCommonArgParser, parseCommonArgs
from presto.Annotation import parseAnnotation, flattenAnnotation, mergeAnnotation
from presto.Applications import runCDHit, runUClust, default_max_memory
from presto.IO import countSeqFile, getFileType, getOutputHandle, printLog, printMessage, \
                      printProgress, readSeqFile, printError, printWarning
from presto.Sequence import indexSeqSets
from presto.Multiprocessing import SeqResult, manageProcesses, feedSeqQueue, \
                                   collectSeqQueue

# Defaults
choices_cluster_tool = ['usearch', 'vsearch', 'cd-hit-est']
default_cluster_tool = 'usearch'
default_cluster_exec = default_usearch_exec
default_cluster_ident = 0.9
default_length_ratio = 0.0
default_cluster_prefix=''
map_cluster_tool = {'cd-hit-est': runCDHit,
                    'usearch': runUClust,
                    'vsearch': runUClust}
min_cluster_ident = {'cd-hit-est': 0.80,
                     'usearch': 0.0,
                     'vsearch': 0.0}

def processQueue(alive, data_queue, result_queue,
                 cluster_func, cluster_args={},
                 cluster_field=default_cluster_field,
                 cluster_prefix=default_cluster_prefix,
                 delimiter=default_delimiter):
    """
    Pulls from data queue, performs calculations, and feeds results queue

    Arguments:
      alive : a multiprocessing.Value boolean controlling whether processing
              continues; when False function returns.
      data_queue : a multiprocessing.Queue holding data to process.
      result_queue : a multiprocessing.Queue to hold processed results.
      cluster_func : the function to use for clustering.
      cluster_args : a dictionary of optional arguments for the clustering function.
      cluster_field : string defining the output cluster field name.
      cluster_prefix : string defining a prefix for the cluster identifier.
      delimiter : a tuple of delimiters for (annotations, field/values, value lists).

    Returns:
      None
    """
    try:
        # Iterator over data queue until sentinel object reached
        while alive.value:
            # Get data from queue
            if data_queue.empty():  continue
            else:  data = data_queue.get()
            # Exit upon reaching sentinel
            if data is None:  break

            # Define result object
            result = SeqResult(data.id, data.data)
            result.log['BARCODE'] = data.id
            result.log['SEQCOUNT'] = len(data)

            # Perform clustering
            cluster_dict = cluster_func(data.data, **cluster_args)

            # Process failed result
            if cluster_dict is None:
                # Update log
                result.log['CLUSTERS'] = 0
                for i, seq in enumerate(data.data, start=1):
                    result.log['CLUST0-%i' % i] = str(seq.seq)

                # Feed results queue and continue
                result_queue.put(result)
                continue

            # Get number of clusters
            result.log['CLUSTERS'] = len(cluster_dict)

            # Update sequence annotations with cluster assignments
            results = list()
            seq_dict = {s.id: s for s in data.data}
            for cluster, id_list in cluster_dict.items():
                for i, seq_id in enumerate(id_list, start=1):
                    # Add cluster annotation
                    seq = seq_dict[seq_id]
                    label = '%s%i' % (cluster_prefix, cluster)
                    header = parseAnnotation(seq.description, delimiter=delimiter)
                    header = mergeAnnotation(header, {cluster_field: label}, delimiter=delimiter)
                    seq.id = seq.name = flattenAnnotation(header, delimiter=delimiter)
                    seq.description = ''

                    # Update log and results
                    result.log['CLUST%i-%i' % (cluster, i)] = str(seq.seq)
                    results.append(seq)

            # Check results
            result.results = results
            result.valid = (len(results) == len(seq_dict))
            # Feed results to result queue
            result_queue.put(result)
        else:
            sys.stderr.write('PID %s> Error in sibling process detected. Cleaning up.\n' \
                             % os.getpid())

            return None
    except:
        alive.value = False
        printError('Error processing sequence set with ID: %s.' % data.id, exit=False)
        raise

    return None


def clusterSets(seq_file, ident=default_cluster_ident, length_ratio=default_length_ratio,
                seq_start=0, seq_end=None, set_field=default_barcode_field, cluster_field=default_cluster_field,
                cluster_prefix=default_cluster_prefix, cluster_memory=default_max_memory,
                cluster_tool=default_cluster_tool, cluster_exec=default_cluster_exec,
                out_file=None, out_args=default_out_args, nproc=None, queue_size=None):
    """
    Performs clustering on sets of sequences

    Arguments:
      seq_file : the sample sequence file name.
      ident : the identity threshold for clustering sequences.
      length_ratio : minimum short/long length ratio allowed within a cluster.
      seq_start : the start position to trim sequences at before clustering.
      seq_end : the end position to trim sequences at before clustering.
      set_field : the annotation containing set IDs.
      cluster_field : the name of the output cluster field.
      cluster_prefix : string defining a prefix for the cluster identifier.
      cluster_memory : cd-hit-est max memory limit (Mb)
      cluster_exec : the path to the clustering executable.
      cluster_tool : the clustering tool to use; one of cd-hit or usearch.
            out_file : output file name. Automatically generated from the input file if None.
      out_file : output file name. Automatically generated from the input file if None.
      out_args : common output argument dictionary from parseCommonArgs.
      nproc : the number of processQueue processes;
              if None defaults to the number of CPUs.
      queue_size : maximum size of the argument queue;
                   if None defaults to 2*nproc.

    Returns:
      str: the clustered output file name.
    """
    # Print parameter info
    log = OrderedDict()
    log['START'] = 'ClusterSets'
    log['COMMAND'] = 'set'
    log['FILE'] = os.path.basename(seq_file)
    log['IDENTITY'] = ident
    log['SEQUENCE_START'] = seq_start
    log['SEQUENCE_END'] = seq_end
    log['SET_FIELD'] = set_field
    log['CLUSTER_FIELD'] = cluster_field
    log['CLUSTER_PREFIX'] = cluster_prefix
    log['CLUSTER_TOOL'] = cluster_tool
    log['NPROC'] = nproc
    printLog(log)

    # Set cluster tool
    try:
        cluster_func = map_cluster_tool.get(cluster_tool)
    except:
        printError('Invalid clustering tool %s.' % cluster_tool)

    # Check the minimum identity
    if ident < min_cluster_ident[cluster_tool]:
        printError('Minimum identity %s too low for clustering tool %s.' % (str(ident), cluster_tool))

    # Define cluster function parameters
    cluster_args = {'cluster_exec': cluster_exec,
                    'ident': ident,
                    'length_ratio': length_ratio,
                    'seq_start': seq_start,
                    'seq_end': seq_end,
                    'max_memory': cluster_memory}

    # Define feeder function and arguments
    index_args = {'field': set_field, 'delimiter': out_args['delimiter']}
    feed_func = feedSeqQueue
    feed_args = {'seq_file': seq_file,
                 'index_func': indexSeqSets,
                 'index_args': index_args}
    # Define worker function and arguments
    work_func = processQueue
    work_args = {'cluster_func': cluster_func,
                 'cluster_args': cluster_args,
                 'cluster_field': cluster_field,
                 'cluster_prefix': cluster_prefix,
                 'delimiter': out_args['delimiter']}
    # Define collector function and arguments
    collect_func = collectSeqQueue
    collect_args = {'seq_file': seq_file,
                    'label': 'cluster',
                    'out_file': out_file,
                    'out_args': out_args,
                    'index_field': set_field}

    # Call process manager
    result = manageProcesses(feed_func, work_func, collect_func,
                             feed_args, work_args, collect_args,
                             nproc, queue_size)

    # Print log
    log = OrderedDict()
    log['OUTPUT'] = result['log'].pop('OUTPUT')
    for k, v in result['log'].items():  log[k] = v
    log['END'] = 'ClusterSets'
    printLog(log)

    return result['out_files']


def clusterAll(seq_file, ident=default_cluster_ident, length_ratio=default_length_ratio,
               seq_start=0, seq_end=None, cluster_field=default_cluster_field,
               cluster_prefix=default_cluster_prefix, cluster_memory=default_max_memory,
               cluster_tool=default_cluster_tool, cluster_exec=default_cluster_exec,
               out_file=None, out_args=default_out_args, nproc=None):
    """
    Performs clustering on all sequences

    Arguments:
      seq_file : the sample sequence file name.
      ident : the identity threshold for clustering sequences.
      length_ratio : minimum short/long length ratio allowed within a cluster.
      seq_start : the start position to trim sequences at before clustering.
      seq_end : the end position to trim sequences at before clustering.
      cluster_field : the name of the output cluster field.
      cluster_prefix : string defining a prefix for the cluster identifier.
      cluster_memory : cd-hit-est max memory limit (Mb)
      cluster_tool : the clustering tool to use; one of cd-hit or usearch.
      cluster_exec : the path to the executable for usearch.
      out_file : output file name. Automatically generated from the input file if None.
      out_args : output arguments.
      nproc : the number of processQueue processes;
              if None defaults to the number of CPUs

    Returns:
      str : the clustered output file name
    """
    # Function to modify SeqRecord header with cluster identifier
    def _header(seq, cluster, field=cluster_field, prefix=cluster_prefix,
                delimiter=out_args['delimiter']):
        label = '%s%i' % (prefix, cluster)
        header = parseAnnotation(seq.description, delimiter=delimiter)
        header = mergeAnnotation(header, {field: label}, delimiter=delimiter)
        seq.id = seq.name = flattenAnnotation(header, delimiter=delimiter)
        seq.description = ''
        return seq

    # Print parameter info
    log = OrderedDict()
    log['START'] = 'ClusterSets'
    log['COMMAND'] = 'all'
    log['FILE'] = os.path.basename(seq_file)
    log['IDENTITY'] = ident
    log['SEQUENCE_START'] = seq_start
    log['SEQUENCE_END'] = seq_end
    log['CLUSTER_FIELD'] = cluster_field
    log['CLUSTER_PREFIX'] = cluster_prefix
    log['CLUSTER_TOOL'] = cluster_tool
    log['NPROC'] = nproc
    printLog(log)

    # Set cluster tool
    try:
        cluster_func = map_cluster_tool.get(cluster_tool)
    except:
        printError('Invalid clustering tool %s.' % cluster_tool)

    # Check the minimum identity
    if ident < min_cluster_ident[cluster_tool]:
        printError('Minimum identity %s too low for clustering tool %s.' % (str(ident), cluster_tool))

    # Count sequence file and parse into a list of SeqRecords
    result_count = countSeqFile(seq_file)
    seq_iter = readSeqFile(seq_file)

    # Perform clustering
    start_time = time()
    printMessage('Running %s' % cluster_tool, start_time=start_time, width=25)
    cluster_dict = cluster_func(seq_iter, ident=ident, length_ratio=length_ratio,
                                seq_start=seq_start, seq_end=seq_end,
                                max_memory=cluster_memory,
                                threads=nproc, cluster_exec=cluster_exec)
    printMessage('Done', start_time=start_time, end=True, width=25)

    # Determine file type
    if out_args['out_type'] is None:
        out_args['out_type'] = getFileType(seq_file)

    # Open output file handles
    if out_file is not None:
        pass_handle = open(out_file, 'w')
    else:
        pass_handle = getOutputHandle(seq_file,
                                      'cluster-pass',
                                      out_dir=out_args['out_dir'],
                                      out_name=out_args['out_name'],
                                      out_type=out_args['out_type'])

    # Open indexed sequence file
    seq_dict = readSeqFile(seq_file, index=True)

    # Iterate over sequence records and update header with cluster annotation
    start_time = time()
    rec_count = pass_count = 0
    for cluster, id_list in cluster_dict.items():
        printProgress(rec_count, result_count, 0.05, start_time=start_time)
        rec_count += len(id_list)

        # Define output sequences
        seq_output = [_header(seq_dict[x], cluster) for x in id_list]

        # Write output
        pass_count += len(seq_output)
        SeqIO.write(seq_output, pass_handle, out_args['out_type'])

    # Update progress
    printProgress(rec_count, result_count, 0.05, start_time=start_time)

    # Print log
    log = OrderedDict()
    log['OUTPUT'] = os.path.basename(pass_handle.name)
    log['CLUSTERS'] = len(cluster_dict)
    log['SEQUENCES'] = result_count
    log['PASS'] = pass_count
    log['FAIL'] = rec_count - pass_count
    log['END'] = 'ClusterSets'
    printLog(log)

    # Close handles
    pass_handle.close()

    return pass_handle.name


def clusterBarcodes(seq_file, ident=default_cluster_ident, length_ratio=default_length_ratio,
                    barcode_field=default_barcode_field, cluster_field=default_cluster_field,
                    cluster_prefix=default_cluster_prefix, cluster_memory=default_max_memory,
                    cluster_tool=default_cluster_tool, cluster_exec=default_cluster_exec,
                    out_file=None, out_args=default_out_args, nproc=None):
    """
    Performs clustering on sets of sequences

    Arguments:
      seq_file : the sample sequence file name.
      ident : the identity threshold for clustering sequences.
      length_ratio : minimum short/long length ratio allowed within a cluster.
      barcode_field : the annotation field containing barcode sequences.
      cluster_field : the name of the output cluster field.
      cluster_prefix : string defining a prefix for the cluster identifier.
      cluster_memory : cd-hit-est max memory limit (Mb)
      seq_start : the start position to trim sequences at before clustering.
      seq_end : the end position to trim sequences at before clustering.
      cluster_tool : the clustering tool to use; one of cd-hit or usearch.
      cluster_exec : the path to the executable for usearch.
      out_file : output file name. Automatically generated from the input file if None.
      out_args : output arguments.
      nproc : the number of processQueue processes;
              if None defaults to the number of CPUs.

    Returns:
      str: the clustered output file name
    """
    # Function to modify SeqRecord header with cluster identifier
    def _header(seq, cluster, field=cluster_field, prefix=cluster_prefix,
                delimiter=out_args['delimiter']):
        label = '%s%i' % (prefix, cluster)
        header = parseAnnotation(seq.description, delimiter=delimiter)
        header = mergeAnnotation(header, {field: label}, delimiter=delimiter)
        seq.id = seq.name = flattenAnnotation(header, delimiter=delimiter)
        seq.description = ''
        return seq

    # Function to extract to make SeqRecord object from a barcode annotation
    def _barcode(seq, field=barcode_field, delimiter=out_args['delimiter']):
        header = parseAnnotation(seq.description, delimiter=delimiter)
        return SeqRecord(Seq(header[field]), id=seq.id)

    # Print parameter info
    log = OrderedDict()
    log['START'] = 'ClusterSets'
    log['COMMAND'] = 'barcode'
    log['FILE'] = os.path.basename(seq_file)
    log['IDENTITY'] = ident
    log['BARCODE_FIELD'] = barcode_field
    log['CLUSTER_FIELD'] = cluster_field
    log['CLUSTER_PREFIX'] = cluster_prefix
    log['CLUSTER_TOOL'] = cluster_tool
    log['NPROC'] = nproc
    printLog(log)

    # Set cluster tool
    try:
        cluster_func = map_cluster_tool.get(cluster_tool)
    except:
        printError('Invalid clustering tool %s.' % cluster_tool)

    # Check the minimum identity
    if ident < min_cluster_ident[cluster_tool]:
        printError('Minimum identity %s too low for clustering tool %s.' % (str(ident), cluster_tool))

    # Count sequence file and parse into a list of SeqRecords
    result_count = countSeqFile(seq_file)
    barcode_iter = (_barcode(x) for x in readSeqFile(seq_file))

    # Perform clustering, note reduced min_word_match compared to cluster sets
    # see https://github.com/torognes/vsearch/issues/328
    start_time = time()
    printMessage('Running %s' % cluster_tool, start_time=start_time, width=25)
    cluster_dict = cluster_func(barcode_iter, ident=ident, length_ratio=length_ratio,
                                seq_start=0, seq_end=None, max_memory=cluster_memory,
                                threads=nproc, cluster_exec=cluster_exec,
                                min_word_match=1)
    printMessage('Done', start_time=start_time, end=True, width=25)

    # Determine file type
    if out_args['out_type'] is None:
        out_args['out_type'] = getFileType(seq_file)

    # Open output file handles
    if out_file is not None:
        pass_handle = open(out_file, 'w')
    else:
        pass_handle = getOutputHandle(seq_file,
                                      'cluster-pass',
                                      out_dir=out_args['out_dir'],
                                      out_name=out_args['out_name'],
                                      out_type=out_args['out_type'])

    # Open indexed sequence file
    seq_dict = readSeqFile(seq_file, index=True)

    # Iterate over sequence records and update header with cluster annotation
    start_time = time()
    rec_count = pass_count = 0
    for cluster, id_list in cluster_dict.items():
        printProgress(rec_count, result_count, 0.05, start_time=start_time)
        rec_count += len(id_list)

        # TODO:  make a generator. Figure out how to get pass_count updated
        # Define output sequences
        seq_output = [_header(seq_dict[x], cluster) for x in id_list]

        # Write output
        pass_count += len(seq_output)
        SeqIO.write(seq_output, pass_handle, out_args['out_type'])

    # Update progress
    printProgress(rec_count, result_count, 0.05, start_time=start_time)

    # Print log
    log = OrderedDict()
    log['OUTPUT'] = os.path.basename(pass_handle.name)
    log['CLUSTERS'] = len(cluster_dict)
    log['SEQUENCES'] = result_count
    log['PASS'] = pass_count
    log['FAIL'] = rec_count - pass_count
    log['END'] = 'ClusterSets'
    printLog(log)

    # Close handles
    pass_handle.close()

    return pass_handle.name


def getArgParser():
    """
    Defines the ArgumentParser

    Returns:
      argparse.ArgumentParser: argument parser object.
    """
    # Define output file names and header fields
    fields = dedent(
             '''
             output files:
                 cluster-pass
                    clustered reads.
                 cluster-fail
                    raw reads failing clustering.

             output annotation fields:
                 CLUSTER
                    a numeric cluster identifier defining the within-group cluster.
             ''')

    # Define ArgumentParser
    parser = ArgumentParser(description=__doc__, epilog=fields,
                            formatter_class=CommonHelpFormatter, add_help=False)
    group_help = parser.add_argument_group('help')
    group_help.add_argument('--version', action='version',
                            version='%(prog)s:' + ' %s %s' %(__version__, __date__))
    group_help.add_argument('-h', '--help', action='help', help='show this help message and exit')
    subparsers = parser.add_subparsers(title='subcommands', dest='command', metavar='',
                                       help='Clustering method')
    # TODO:  This is a temporary fix for Python issue 9253
    subparsers.required = True

    # Parent parser of common arguments
    parent_parser = ArgumentParser(formatter_class=CommonHelpFormatter, add_help=False)
    group_parent = parent_parser.add_argument_group('common clustering arguments')
    group_parent.add_argument('-k', action='store', dest='cluster_field', type=str,
                              default=default_cluster_field,
                              help='''The name of the output annotation field to add with the
                                   cluster information for each sequence.''')
    group_parent.add_argument('--ident', action='store', dest='ident', type=float,
                              default=default_cluster_ident,
                              help='''The sequence identity threshold to use for clustering. 
                                   Note, how identity is calculated is specific to the clustering 
                                   application used.''')
    group_parent.add_argument('--length', action='store', dest='length_ratio', type=float,
                              default=default_length_ratio,
                              help='''The minimum allowed shorter/longer sequence length ratio allowed 
                                   within a cluster. Setting this value to 1.0 will require identical 
                                   length matches within clusters. A value of 0.0 will allow clusters
                                   containing any length of substring.''')
    group_parent.add_argument('--prefix', action='store', dest='cluster_prefix', type=str,
                              default=default_cluster_prefix,
                               help='''A string to use as the prefix for each cluster identifier.
                                    By default, cluster identifiers will be numeric values only.''')
    group_parent.add_argument('--cluster', action='store', dest='cluster_tool',
                              choices=choices_cluster_tool, default=default_cluster_tool,
                              help='''The clustering tool to use for assigning clusters. 
                                   Must be one of usearch, vsearch or cd-hit-est. Note, for 
                                   cd-hit-est the default maximum memory limit is set to 3GB.''')
    group_parent.add_argument('--mem', action='store', dest='cluster_memory',
                              default=default_max_memory,
                              help='''The maximum memory limit for cd-hit-est in MB. 
                                   Ignored if using usearch or vsearch.''')
    group_parent.add_argument('--exec', action='store', dest='cluster_exec', default=None,
                              help='The name or path of the usearch, vsearch or cd-hit-est executable.')

    # Sequence set clustering arguments
    parser_set = subparsers.add_parser('set',
                                       parents=[getCommonArgParser(log=True, multiproc=True), parent_parser],
                                       formatter_class=CommonHelpFormatter, add_help=False,
                                       help='Cluster sequences within annotation sets.',
                                       description='Cluster sequences within annotation sets.')
    group_set = parser_set.add_argument_group('grouped sequence clustering arguments')
    group_set.add_argument('-f', action='store', dest='set_field', type=str,
                           default=default_barcode_field,
                           help='''The annotation field containing annotations, such as UMI
                                barcode, for sequence grouping.''')
    group_set.add_argument('--start', action='store', dest='seq_start', type=int, default=0,
                           help='''The start of the region to be used for clustering.
                                Together with --end, this parameter can be used to specify a
                                subsequence of each read to use in the clustering algorithm.''')
    group_set.add_argument('--end', action='store', dest='seq_end', type=int,
                           help='The end of the region to be used for clustering.')
    parser_set.set_defaults(func=clusterSets)

    # Total sequence clustering arguments
    parser_all = subparsers.add_parser('all',
                                       parents=[getCommonArgParser(log=False, failed=False, multiproc=True), parent_parser],
                                       formatter_class=CommonHelpFormatter, add_help=False,
                                       help='Cluster all sequences regardless of annotation.',
                                       description='Cluster all sequences regardless of annotation.')
    group_all = parser_all.add_argument_group('total sequence clustering arguments')
    group_all.add_argument('--start', action='store', dest='seq_start', type=int,
                           help='''The start of the region to be used for clustering.
                                Together with --end, this parameter can be used to specify a
                                subsequence of each read to use in the clustering algorithm.''')
    group_all.add_argument('--end', action='store', dest='seq_end', type=int,
                           help='The end of the region to be used for clustering.')
    parser_all.set_defaults(func=clusterAll)

    # Sequence set clustering arguments
    parser_barcode = subparsers.add_parser('barcode',
                                           parents=[getCommonArgParser(log=False, failed=False, multiproc=True), parent_parser],
                                           formatter_class=CommonHelpFormatter, add_help=False,
                                           help='Cluster reads by clustering barcode sequences.',
                                           description='Cluster reads by clustering barcode sequences.')
    group_barcode = parser_barcode.add_argument_group('barcode clustering arguments')
    group_barcode.add_argument('-f', action='store', dest='barcode_field', type=str,
                               default=default_barcode_field,
                               help='''The annotation field containing barcode sequences.''')
    parser_barcode.set_defaults(func=clusterBarcodes)

    return parser


if __name__ == '__main__':
    """
    Parses command line arguments and calls main function
    """
    # Parse arguments
    parser = getArgParser()
    checkArgs(parser)
    args = parser.parse_args()
    args_dict = parseCommonArgs(args)

    # Convert fields to uppercase
    if 'set_field' in args_dict and args_dict['set_field'] is not None:
        args_dict['set_field'] = args_dict['set_field'].upper()
    if 'cluster_field' in args_dict and args_dict['cluster_field'] is not None:
        args_dict['cluster_field'] = args_dict['cluster_field'].upper()

    # Set cluster exec if unspecified
    if args_dict['cluster_exec'] is None:
        args_dict['cluster_exec'] = args_dict['cluster_tool']

    # Check if a valid clustering executable was specified
    if not shutil.which(args_dict['cluster_exec']):
        parser.error('%s executable not found' % args_dict['cluster_exec'])

    # Check for valid start and end input
    if ('seq_start' in args_dict and 'seq_end' in args_dict) and \
            args_dict['seq_start'] is not None and args_dict['seq_end'] is not None and \
            args_dict['seq_start'] >= args_dict['seq_end']:
        parser.error('--start must be less than --end')

    # Call cluster main function for each input file
    del args_dict['seq_files']
    del args_dict['func']
    del args_dict['command']
    if 'out_files' in args_dict:  del args_dict['out_files']
    for i, f in enumerate(args.__dict__['seq_files']):
        args_dict['seq_file'] = f
        args_dict['out_file'] = args.__dict__['out_files'][i] \
            if args.__dict__['out_files'] else None
        args.func(**args_dict)

    # import cProfile
    # prof = cProfile.Profile()
    # results = prof.runcall(args.func, **args_dict)
    # prof.dump_stats('cluster.prof')



"""
Annotation functions
"""
# Info
__author__ = 'Jason Anthony Vander Heiden'
from presto import __version__, __date__

# Imports
import re
from collections import OrderedDict
from presto.Defaults import default_coord, default_delimiter, default_separator


def getCoordKey(header, coord_type=default_coord, delimiter=default_delimiter):
    """
    Return the coordinate identifier for a sequence description

    Arguments:
      header (str): Sequence header string
      coord_type (str): Sequence header format;
                   one of 'illumina', 'solexa', 'sra', 'ena', '454', or 'presto';
                   if unrecognized type or None, then return the input header.
      delimiter (tuple): Tuple of delimiters for (fields, values, value lists)

    Returns:
      str: Coordinate identifier as a string.
    """
    if coord_type in ('illumina', 'solexa'):
        return header.split()[0].split('/')[0].split('#')[0]
    elif coord_type in ('sra', 'ena'):
        return '.'.join(header.split()[0].split('.')[:2])
    elif coord_type == '454':
        return header.split()[0]
    elif coord_type == 'presto':
        return parseAnnotation(header, delimiter=delimiter)['ID']
    else:
        return header


def parseLog(record):
    """
    Parses an pRESTO log record

    Arguments:
      record (str): a string of lines representing a log record including newline characters.

    Returns:
      collections.OrderedDict: parsed log contain field and values pairs as a dictionary.
    """
    record_dict = OrderedDict()
    for line in record.split('\n'):
        line = line.strip().split('> ')
        if len(line) == 2:
            record_dict[line[0]] = line[1]

    return record_dict


def parseAnnotation(record, fields=None, delimiter=default_delimiter):
    """
    Extracts annotations from a FASTA/FASTQ sequence description

    Arguments:
      record : Description string to extract annotations from
      fields : List of fields to subset the return dictionary to;
               if None return all fields
      delimiter : a tuple of delimiters for (fields, values, value lists)

    Returns:
      OrderedDict : An OrderedDict of field/value pairs
    """
    annotation = record.split(delimiter[0])
    field_dict = OrderedDict([('ID', annotation.pop(0))])
    for ann in annotation:
        vals = ann.split(delimiter[1])
        field_dict[vals[0].upper()] = vals[1]

    # Subset return dictionary to requested fields
    if fields is not None:
        if not isinstance(fields, list):  fields = [fields]
        for f in set(field_dict).difference(fields):  del field_dict[f]

    return field_dict


def flattenAnnotation(ann_dict, delimiter=default_delimiter):
    """
    Converts annotations from a dictionary to a FASTA/FASTQ sequence description

    Arguments:
      ann_dict : Dictionary of field/value pairs
      delimiter : Tuple of delimiters for (fields, values, value lists)

    Returns:
      str : Formatted sequence description string
    """
    annotation = ann_dict.get('ID', 'NONE')
    for k, v in ann_dict.items():
        # Skip ID field
        if k.upper() == 'ID':
            continue

        if isinstance(v, list):
            v = delimiter[2].join([str(x) for x in v])
        annotation += '%s%s%s%s' % (delimiter[0], k.upper(), delimiter[1], v)

    return annotation


def mergeAnnotation(ann_dict_1, ann_dict_2, prepend=False,
                    delimiter=default_delimiter):
    """
    Merges non-ID field annotations from one field dictionary into another

    Arguments:
      ann_dict_1 : Dictionary of field/value pairs to append to
      ann_dict_2 : Dictionary of field/value pairs to merge with ann_dict_2
      prepend : If True then add ann_dict_2 values to the front of any ann_dict_1
                values that are already present, rather than the default behavior
                of appending ann_dict_2 values.
      delimiter : Tuple of delimiters for (fields, values, value lists)

    Returns:
      OrderedDict : Modified ann_dict_1 dictonary of field/value pairs
    """
    # Define merge order
    if prepend:
        def _merge(x, y):  return '%s%s%s' % (y, delimiter[2], x)
    else:
        def _merge(x, y):  return '%s%s%s' % (x, delimiter[2], y)

    merged_dict = ann_dict_1.copy()
    for k, v in ann_dict_2.items():
        # Skip ID field
        if k.upper() == 'ID':
            continue

        if k in merged_dict:
            if isinstance(merged_dict[k], list):
                merged_dict[k] = delimiter[2].join([str(x) for x in merged_dict[k]])
            if isinstance(v, list):
                v = delimiter[2].join([str(x) for x in v])
            merged_dict[k] = _merge(merged_dict[k], v)
        else:
            merged_dict[k.upper()] = v

    return merged_dict


def renameAnnotation(ann_dict, old_field, new_field, delimiter=default_delimiter):
    """
    Renames an annotation and merges annotations if the new name already exists

    Arguments:
      ann_dict : Dictionary of field/value pairs
      old_field : Old field name
      new_field : New field name
      delimiter : Tuple of delimiters for (fields, values, value lists)

    Returns:
      OrderedDict : Modified fields dictonary
    """
    if new_field in ann_dict:
        rename_dict = ann_dict.copy()
        del rename_dict[old_field]
        rename_dict = mergeAnnotation(rename_dict, {new_field:ann_dict[old_field]},
                                      delimiter=delimiter)
    else:
        rename_dict = OrderedDict([(new_field, v) if k == old_field else (k, v) \
                                   for k, v in ann_dict.items()])

    return rename_dict


# TODO:  this converted min/max/sum collapse to strings instead of floats (for rounding purposes). which is odd.
def collapseAnnotation(ann_dict, action, fields=None, delimiter=default_delimiter):
    """
    Collapses multiple annotations into new single annotations for each field

    Arguments:
      ann_dict : dictionary of field/value pairs
      action : collapse action to take;
               one of {min, max, sum, first, last, set, cat}
      fields : subset of ann_dict to _collapse;
               if None, collapse all but the ID field
      delimiter : Tuple of delimiters for (fields, values, value lists)

    Returns:
      OrderedDict : Modified field dictionary
    """
    # Define _collapse action
    if action == 'set':
        def _collapse(value):  return sorted(set(value))
    elif action == 'first':
        def _collapse(value):  return value[0]
    elif action == 'last':
        def _collapse(value):  return value[-1]
    elif action == 'min':
        def _collapse(value):  return '%.12g' % min([float(x or 0) for x in value])
    elif action == 'max':
        def _collapse(value):  return '%.12g' % max([float(x or 0) for x in value])
    elif action == 'sum':
        def _collapse(value):  return '%.12g' % sum([float(x or 0) for x in value])
    elif action == 'cat':
        def _collapse(value):  return ''.join([str(x) for x in value])
    else:
        def _collapse(value):  return value

    # Collapse fields
    collapse_dict = ann_dict.copy()
    for k, v in collapse_dict.items():
        if k.upper() == 'ID':
            continue
        if fields is None or k in fields:
            # Convert field to list
            if not isinstance(v, list) and isinstance(v, str):
                v = v.split(delimiter[2])
            elif not isinstance(v, list):
                v = [v]
            # Perform _collapse and reassign field
            collapse_dict[k] = _collapse(v)

    return collapse_dict


def getAnnotationValues(seq_iter, field, unique=False, delimiter=default_delimiter):
    """
    Gets the set of unique annotation values in a sequence set

    Arguments:
      seq_iter : Iterator or list of SeqRecord objects
      field : Annotation field to retrieve values for
      unique : If True return a list of only the unique values;
               if False return a list of all values
      delimiter : Tuple of delimiters for (fields, values, value lists)

    Returns:
      list : List of values for the field
    """
    # Parse annotations from seq_list records
    ann_iter = (parseAnnotation(s.description, delimiter=delimiter) for s in seq_iter)
    values = [a[field] for a in ann_iter]

    return list(set(values)) if unique else values


def annotationConsensus(seq_iter, field, delimiter=default_delimiter):
    """
    Calculate a consensus annotation for a set of sequences

    Arguments:
      seq_iter : an iterator or list of SeqRecord objects
      field : the annotation field to take a consensus of
      delimiter : a tuple of delimiters for (annotations, field/values, value lists)

    Returns:
      dict : Dictionary with keys
             `set` containing a list of unique annotation values,
             `count` containing annotation counts,
             `cons` containing the consensus annotation,
             `freq` containing the majority annotation frequency
    """
    # Define return dictionary
    cons_dict = {'set':None, 'count':None, 'cons':None, 'freq':None}

    # Parse annotations from seq_list records
    val_list = getAnnotationValues(seq_iter, field, delimiter=delimiter)

    # Define annotation set and counts
    cons_dict['set'] = sorted(set(val_list))
    cons_dict['count'] = [val_list.count(v) for v in cons_dict['set']]

    # Define consensus annotation
    i = cons_dict['count'].index(max(cons_dict['count']))
    cons_dict['cons'] = cons_dict['set'][i]
    cons_dict['freq'] = float(cons_dict['count'][i]) / len(val_list)

    return cons_dict


def convertGenericHeader(desc, delimiter=default_delimiter):
    """
    Converts any header to the pRESTO format

    Arguments:
      desc (str): a sequence description string.
      delimiter (tuple): a tuple of delimiters for (fields, values, value lists).

    Returns:
      dict: a dictionary of header field and value pairs.
    """
    # Replace whitespace and delimiter characters
    sub_regex = '[%s\s]+' % re.escape(''.join(delimiter))
    conv = re.sub(sub_regex, '_', desc)
    try:
        # Check if modified header is valid
        header = parseAnnotation(conv, delimiter=delimiter)
    except:
        # Assign header to None if header cannot be converted
        header = None

    return header


def convert454Header(desc):
    """
    Parses 454 headers into the pRESTO format

    Arguments:
      desc (str): a sequence description string.

    Returns:
      dict: a dictionary of header field and value pairs.

    Examples:
      New style 454 header::

        @<accession> <length=##>
        @GXGJ56Z01AE06X length=222

      Old style 454 header::

        @<rank_x_y> <length=##> <uaccno=accession>
        @000034_0199_0169 length=437 uaccno=GNDG01201ARRCR
    """
    # Split description and assign field names
    try:
        # Build header dictionary
        fields = desc.split(' ')
        header = OrderedDict()
        header['ID'] = fields[0]
        header['LENGTH'] = fields[1].replace('length=', '')

        # Check for old format
        if len(fields) == 3:
            header['UACCNO'] = fields[2].replace('uaccno=', '')
        elif len(fields) != 2:
            raise ValueError
    except:
        header = None

    return header


def convertGenbankHeader(desc, delimiter=default_delimiter):
    """
    Converts GenBank and RefSeq headers into the pRESTO format

    Arguments:
      desc (str): a sequence description string.
      delimiter (tuple): a tuple of delimiters for (fields, values, value lists).

    Returns:
      dict: a dictionary of header field and value pairs.

    Examples:
      New style GenBank header::

        <accession>.<version> <description>
        >CM000663.2 Homo sapiens chromosome 1, GRCh38 reference primary assembly

      Old style GenBank header::

        gi|<GI record number>|<dbsrc>|<accession>.<version>|<description>
        >gi|568336023|gb|CM000663.2| Homo sapiens chromosome 1, GRCh38 reference primary assembly
    """
    # Define special characters to replace
    sub_regex = '[%s\s]+' % re.escape(''.join(delimiter[1:]))

    # Split description and assign field names
    try:
        header = OrderedDict()

        # Try old format and fallback to new format if that fails
        fields = desc.split('|')
        if len(fields) == 5:
            header['ID'] = fields[3]
            header['GI'] = fields[1]
            header['SOURCE'] = fields[2]
            header['DESC'] = re.sub(sub_regex, '_', fields[4].strip())
        else:
            fields = desc.split(' ')
            header['ID'] = fields[0]
            header['DESC'] = re.sub(sub_regex, '_', '_'.join(fields[1:]).strip())
    except:
        header = None

    return header


def convertIlluminaHeader(desc):
    """
    Converts Illumina headers into the pRESTO format

    Arguments:
      desc (str): a sequence description string.

    Returns:
      dict: a dictionary of header field and value pairs.

    Examples:
      New style Illumina header::

        @<instrument>:<run number>:<flowcell ID>:<lane>:<tile>:<x-pos>:<y-pos> <read number>:<is filtered>:<control number>:<index sequence>
        @MISEQ:132:000000000-A2F3U:1:1101:14340:1555 2:N:0:ATCACG

      Old style Illumina header::

        @<instrument>:<flowcell lane>:<tile>:<x-pos>:<y-pos>#<index sequence>/<read number>
        @HWI-EAS209_0006_FC706VJ:5:58:5894:21141#ATCACG/1
        @MS6_33112:1:1101:18371:1066/1
    """
    # Split description and assign field names
    try:
        # Try new format and fallback to old if that fails
        fields = desc.split(' ')
        if len(fields) == 2:
            x = fields[1].split(':')
            index = x[3]
            read_num = x[0]
        elif '#' in desc:
            fields = desc.split('#')
            x = fields[1].split('/')
            index = x[0]
            read_num = x[1]
        else:
            fields = desc.split('/')
            index = None
            read_num = fields[1]

        # Build header dictionary
        header = OrderedDict()
        header['ID'] = fields[0]
        if index is not None:  header['INDEX'] = index
        header['READ'] = read_num
    except:
        header = None

    return header


def convertIMGTHeader(desc, simple=False):
    """
    Converts germline headers from IMGT/GENE-DB into the pRESTO format

    Arguments:
      desc (str): a sequence description string.
      simple (bool): if True then the header will be converted to only the allele name.

    Returns:
      dict: a dictionary of header field and value pairs.

    Examples:
      IMGT header::

        >X60503|IGHV1-18*02|Homo sapiens|F|V-REGION|142..417|276 nt|1| | | | |276+24=300|partial in 3'| |

      Header contains 15 fields separated by ``|`` (http://imgt.org/genedb):

        1. IMGT/LIGM-DB accession number(s).
        2. Gene and allele name.
        3. Species.
        4. Functionality.
        5. Exon(s), region name(s), or extracted label(s).
        6. Start and end positions in the IMGT/LIGM-DB accession number(s).
        7. Number of nucleotides in the IMGT/LIGM-DB accession number(s).
        8. Codon start, or 'NR' (not relevant) for non coding labels and
           out-of-frame pseudogenes.
        9. Number of nucleotides added in ``5'`` compared to the
           corresponding label extracted from IMGT/LIGM-DB.
        10. Number of nucleotides added or removed in ``3'``
            compared to the corresponding label extracted from IMGT/LIGM-DB.
        11. Number of added, deleted, and/or substituted
            nucleotides to correct sequencing errors, or 'not corrected' if
            non corrected sequencing errors.
        12. Number of amino acids (AA). This field indicates that the
            sequence is in amino acids.
        13. Number of characters in the sequence. Nucleotides (or AA) plus IMGT gaps.
        14. Partial (if it is).
        15. Reverse complementary (if it is).
    """
    # Split description and assign field names
    try:
        fields = desc.split('|')

        # Build header dictionary
        header = OrderedDict()
        header['ID'] = fields[1]

        if not simple:
            header['SPECIES'] = re.sub('\s', '_', fields[2])
            header['REGION'] = fields[4]
            header['FUNCTIONALITY'] = re.sub('[\(\)\[\]]', '', fields[3])
            header['PARTIAL'] = 'FALSE' if re.sub('\s', '', fields[13]) == '' else 'TRUE'
            header['ACCESSION'] = fields[0]

        # Position and length data
        #header['NUCLEOTIDES'] = re.sub('[^0-9]', '', fields[6])
        #header['LENGTH'] = fields[12].split('=')[1]
    except:
        header = None

    return header


def convertSRAHeader(desc):
    """
    Parses NCBI SRA or EMBL-EBI ENA headers into the pRESTO format

    Arguments:
      desc (str): a sequence description string.

    Returns:
      dict: a dictionary of header field and value pairs.

    Examples:
      Header from ``fastq-dump --split-files``::

        @<accession>.<spot> <original sequence description> <length=#>
        @SRR001666.1 071112_SLXA-EAS1_s_7:5:1:817:345 length=36
        @SRR1383326.1 1 length=250

      Header from ``fastq-dump --split-files -I``::

        @<accession>.<spot>.<read number> <original sequence description> <length=#>
        @SRR1383326.1.1 1 length=250

      Header from ENA::

        @<accession>.<spot> <original sequence description>
        @ERR220397.1 HKSQ1MM01DXT2W/3
        @ERR346596.1 BS-DSFCONTROL04:4:000000000-A3F0Y:1:1101:12758:1640/1
        @ERR346596.1 BS-DSFCONTROL04:4:000000000-A3F0Y:1:1101:12758:1640/2
    """
    # Split description and assign field names
    try:
        fields = desc.split(' ')

        # Build header dictionary
        header = OrderedDict()

        # Check for read number if sequence id
        read_id = fields[0].split('.')
        if len(read_id) == 3:
            header['ID'] = '.'.join(read_id[:2])
            header['READ'] = read_id[2]
        else:
            header['ID'] = fields[0]

        header['DESC'] = fields[1]
        if len(fields) >= 3 and 'length' in fields[2]:
            header['LENGTH'] = fields[2].replace('length=', '')
    except:
        header = None

    return header


def convertMIGECHeader(desc):
    """
    Parses headers from the MIGEC tool into the pRESTO format

    Arguments:
      desc (str): a sequence description string.

    Returns:
      dict: a dictionary of header field and value pairs.

    Examples:
      MIGEC header::

        @MIG UMI:<UMI sequence>:<consensus read count>
        @MIG UMI:TCGGCCAACAAA:8
    """
    # Split description and assign field names
    try:
        fields = desc.split(':')

        # Build header dictionary
        header = OrderedDict()
        header['ID'] = fields[1]
        header['COUNT'] = fields[2]
    except:
        header = None

    return header


def addHeader(header, fields, values, delimiter=default_delimiter):
    """
    Adds fields and values to a sequence header

    Arguments:
      header : an annotation dictionary returned by parseAnnotation.
      fields : the list of fields to add or append to.
      values : the list of annotation values to add for each field.
      delimiter : a tuple of delimiters for (fields, values, value lists).

    Returns:
      dict: modified header dictionary.
    """
    for f, v in zip(fields, values):
        header = mergeAnnotation(header, {f:v}, delimiter=delimiter)

    return header


def collapseHeader(header, fields, actions, delimiter=default_delimiter):
    """
    Collapses a sequence header

    Arguments:
      header : an annotation dictionary returned by parseAnnotation.
      fields : the list of fields to collapse.
      actions : the list of collapse action take;
                one of (max, min, sum, first, last, set, cat) for each field.
      delimiter : a tuple of delimiters for (fields, values, value lists).

    Returns:
      dict: modified header dictionary.
    """
    for f, a in zip(fields, actions):
        header = collapseAnnotation(header, a, f, delimiter=delimiter)

    return header


def copyHeader(header, fields, names, actions=None, delimiter=default_delimiter):
    """
    Copies fields in a sequence header

    Arguments:
      header : an annotation dictionary returned by parseAnnotation.
      fields : a list of the field names to copy.
      names : a list of the new field names.
      actions : the list of collapse action take after the copy;
                one of (max, min, sum, first, last, set, cat) for each field.
      delimiter : a tuple of delimiters for (fields, values, value lists).

    Returns:
      dict: modified header dictionary.
    """
    old_header = header.copy()
    for f, n in zip(fields, names):
        header = mergeAnnotation(header, {n: old_header[f]}, delimiter=delimiter)

    if actions is not None:
        header = collapseHeader(header, names, actions, delimiter=delimiter)

    return header


def deleteHeader(header, fields, delimiter=default_delimiter):
    """
    Deletes fields from a sequence header

    Arguments:
      header : an annotation dictionary returned by parseAnnotation.
      fields : the list of fields to delete.
      delimiter : a tuple of delimiters for (fields, values, value lists).

    Returns:
      dict: modified header dictionary
    """
    for f in fields:  del header[f]

    return header


def expandHeader(header, fields, separator=default_separator,
                 delimiter=default_delimiter):
    """
    Splits and annotation value into separate fields in a sequence header

    Arguments:
      header : an annotation dictionary returned by parseAnnotation.
      fields : the field to split.
      separator : the delimiter to split the values by.
      delimiter : a tuple of delimiters for (fields, values, value lists).

    Returns:
      dict: modified header dictionary.
    """
    for f in fields:
        values = header[f].split(separator)
        names = [f + str(i + 1) for i in range(len(values))]
        ann = OrderedDict([(n, v) for n, v in zip(names, values)])
        header = mergeAnnotation(header, ann, delimiter=delimiter)
        del header[f]

    return header


def mergeHeader(header, fields, name, action=None, delete=False,
                delimiter=default_delimiter):
    """
    Merges fields in a sequence header

    Arguments:
      header : an annotation dictionary returned by parseAnnotation.
      fields : a list of the field names to merge.
      name : the name of the new field.
      delete : if True delete the merged fields.
      actions : the list of collapse action take after the merge
                one of (max, min, sum, first, last, set, cat).
      delimiter : a tuple of delimiters for (fields, values, value lists)

    Returns:
      dict: modified header dictionary.
    """
    merge = {name: [header[f] for f in fields]}
    header = mergeAnnotation(header, merge, delimiter=delimiter)

    # Delete fields
    if delete:
        for f in fields:
            if f != name:  del header[f]

    # Collapse action
    if action is not None:
        header = collapseHeader(header, fields=[name], actions=[action],
                                delimiter=delimiter)

    return header


def renameHeader(header, fields, names, actions=None, delimiter=default_delimiter):
    """
    Renames fields in a sequence header

    Arguments:
      header : an annotation dictionary returned by parseAnnotation.
      fields : a list of the current field names.
      names : a list of the new field names.
      actions : the list of collapse action take after the rename;
              one of (max, min, sum, first, last, set, cat) for each field.
      delimiter : a tuple of delimiters for (fields, values, value lists).

    Returns:
      dict: modified header dictionary.
    """
    for f, n in zip(fields, names):
        header = renameAnnotation(header, f, n, delimiter=delimiter)

    if actions is not None:
        header = collapseHeader(header, names, actions, delimiter=delimiter)

    return header



"""
Version and authorship information
"""

__author__    = 'Jason Anthony Vander Heiden'
__copyright__ = 'Copyright 2025 Kleinstein Lab, Yale University. All rights reserved.'
__license__   = 'GNU Affero General Public License 3 (AGPL-3)'
__version__   = '0.7.6'
__date__      = '2025.08.28'



#!/usr/bin/env python3
"""
Removes primers and annotates sequences with primer and barcode identifiers
"""
# Info
__author__ = 'Jason Anthony Vander Heiden'
from presto import __version__, __date__

# Imports
import os
from argparse import ArgumentParser
from collections import OrderedDict
from textwrap import dedent

# Presto imports
from presto.Defaults import default_delimiter, default_out_args, default_primer_gap_penalty, default_primer_max_error, \
                            default_primer_max_len, default_primer_start, default_barcode_field, default_primer_field
from presto.Commandline import CommonHelpFormatter, checkArgs, getCommonArgParser, parseCommonArgs
from presto.Sequence import localAlignment, compilePrimers, extractAlignment, getDNAScoreDict, \
                            maskSeq, reverseComplement, scoreAlignment
from presto.IO import readPrimerFile, printLog
from presto.Multiprocessing import SeqResult, manageProcesses, feedSeqQueue, \
                                   processSeqQueue, collectSeqQueue

def extractPrimers(data, start, length, rev_primer=False, mode='mask', barcode=False,
                   barcode_length=None, barcode_field=default_barcode_field, primer_field=default_primer_field,
                   delimiter=default_delimiter):
    """
    Extracts primer sequences directly

    Arguments:
      data : SeqData object containing a single SeqRecord object to process.
      start : position where subsequence starts.
      length : the length of the subsequence to extract.
      rev_primer : if True extract relative to the tail end of the sequence.
      mode : defines the action taken; one of 'cut', 'mask', 'tag' or 'trim'.
      barcode : if True add sequence preceding primer to description.
      barcode_length : length of the barcode sequence. If None (default), the barcode
                       sequence is the full sequence preceding the primer sequence.
      barcode_field : name of the output barcode annotation.
      primer_field : name of the output primer annotation.
      delimiter : a tuple of delimiters for (annotations, field/values, value lists).

    Returns:
      presto.Multiprocessing.SeqResult: result object.
    """

    # Define result object
    result = SeqResult(data.id, data.data)

    # Align primers
    align = extractAlignment(data.data, start=start, length=length, rev_primer=rev_primer)
    if not align:
        result.log['ALIGN'] = None
        return result

    # Create output sequence
    out_seq = maskSeq(align, mode=mode, barcode=barcode, barcode_length=barcode_length, barcode_field=barcode_field,
                      primer_field=primer_field, delimiter=delimiter)
    result.results = out_seq
    result.valid = True

    # Update log with successful alignment results
    result.log['PRIMER'] = align.primer
    result.log['PRSTART'] = align.start
    if 'barcode' in out_seq.annotations:
        result.log['BARCODE'] = out_seq.annotations['barcode']
    if not align.rev_primer:
        align_cut = len(align.align_seq) - align.gaps
        result.log['INSEQ'] = align.align_seq + str(align.seq.seq[align_cut:])
        result.log['ALIGN'] = align.align_primer
        result.log['OUTSEQ'] = str(out_seq.seq).rjust(len(result.data.seq) + align.gaps)
    else:
        align_cut = len(align.seq) - len(align.align_seq) + align.gaps
        result.log['INSEQ'] = str(align.seq.seq[:align_cut]) + align.align_seq
        result.log['ALIGN'] = align.align_primer.rjust(len(result.data.seq) + align.gaps)
        result.log['OUTSEQ'] = str(out_seq.seq)

    return result


def alignPrimers(data, primers, primers_regex=None, max_error=default_primer_max_error,
                 max_len=default_primer_max_len, rev_primer=False, skip_rc=False, mode='mask',
                 barcode=False, barcode_length=None, barcode_field=default_barcode_field, primer_field=default_primer_field,
                 gap_penalty=default_primer_gap_penalty, score_dict=getDNAScoreDict(mask_score=(0, 1), gap_score=(0, 0)),
                 delimiter=default_delimiter):
    """
    Performs pairwise local alignment of a list of short sequences against a long sequence

    Arguments:
      data : SeqData object containing a single SeqRecord object to process.
      primers : dictionary of {names: short IUPAC ambiguous sequence strings}.
      primers_regex : optional dictionary of {names: compiled primer regular expressions}.
      max_error : maximum acceptable error rate for a valid alignment.
      max_len : maximum length of sample sequence to align.
      rev_primer : if True align with the tail end of the sequence.
      skip_rc : if True do not check reverse complement sequences.
      mode : defines the action taken; one of 'cut', 'mask', 'tag' or 'trim'.
      barcode : if True add sequence preceding primer to description.
      barcode_length : length of the barcode sequence. If None (default), the barcode
                       sequence is the full sequence preceding the primer sequence.
      barcode_field : name of the output barcode annotation.
      primer_field : name of the output primer annotation.
      gap_penalty : a tuple of positive (gap open, gap extend) penalties.
      score_dict : optional dictionary of {(char1, char2): score} alignment scores
      delimiter : a tuple of delimiters for (annotations, field/values, value lists).

    Returns:
      presto.Multiprocessing.SeqResult: result object.
    """
    # Define result object
    result = SeqResult(data.id, data.data)

    # Align primers
    align = localAlignment(data.data, primers, primers_regex=primers_regex, max_error=max_error,
                           max_len=max_len, rev_primer=rev_primer, skip_rc=skip_rc,
                           gap_penalty=gap_penalty, score_dict=score_dict)
    if not align:
        # Update log if no alignment
        result.log['ALIGN'] = None
        return result

    # Create output sequence
    out_seq = maskSeq(align, mode=mode, barcode=barcode, barcode_length=barcode_length, barcode_field=barcode_field,
                      primer_field=primer_field, delimiter=delimiter)
    result.results = out_seq
    result.valid = bool(align.error <= max_error) if len(out_seq) > 0 else False

    # Update log with successful alignment results
    result.log['SEQORIENT'] = out_seq.annotations['seqorient']
    result.log['PRIMER'] = align.primer
    result.log['PRORIENT'] = 'RC' if align.rev_primer else 'F'
    result.log['PRSTART'] = align.start
    if 'barcode' in out_seq.annotations:
        result.log['BARCODE'] = out_seq.annotations['barcode']
    if not align.rev_primer:
        align_cut = len(align.align_seq) - align.gaps
        result.log['INSEQ'] = align.align_seq + \
                              str(align.seq.seq[align_cut:])
        result.log['ALIGN'] = align.align_primer
        result.log['OUTSEQ'] = str(out_seq.seq).rjust(len(result.data.seq) + align.gaps)
    else:
        align_cut = len(align.seq) - len(align.align_seq) + align.gaps
        result.log['INSEQ'] = str(align.seq.seq[:align_cut]) + align.align_seq
        result.log['ALIGN'] = align.align_primer.rjust(len(result.data.seq) + align.gaps)
        result.log['OUTSEQ'] = str(out_seq.seq)
    result.log['ERROR'] = align.error

    return result


def scorePrimers(data, primers, max_error=default_primer_max_error, start=default_primer_start, rev_primer=False, mode='mask',
                 barcode=False, barcode_length=None, barcode_field=default_barcode_field, primer_field=default_primer_field,
                 score_dict=getDNAScoreDict(mask_score=(0, 1), gap_score=(0, 0)),
                 delimiter=default_delimiter):
    """
    Performs a simple fixed position alignment of primers

    Arguments:
      data : SeqData object containing a single SeqRecord object to process.
      primers : dictionary of {names: short IUPAC ambiguous sequence strings}.
      max_error : maximum acceptable error rate for a valid alignment
      start : position where primer alignment starts.
      rev_primer : if True align with the tail end of the sequence.
      mode : defines the action taken; one of 'cut', 'mask', 'tag' or 'trim'.
      barcode : if True add sequence preceding primer to description.
      barcode_length : length of the barcode sequence. If None (default), the barcode
                       sequence is the full sequence preceding the primer sequence.
      barcode_field : name of the output barcode annotation.
      primer_field : name of the output primer annotation.
      score_dict : optional dictionary of {(char1, char2): score} alignment scores
      delimiter : a tuple of delimiters for (annotations, field/values, value lists).

    Returns:
      presto.Multiprocessing.SeqResult: result object.
    """
    # Define result object
    result = SeqResult(data.id, data.data)

    # Align primers
    align = scoreAlignment(data.data, primers, start=start, rev_primer=rev_primer,
                           score_dict=score_dict)
    if not align:
        # Update log if no alignment
        result.log['ALIGN'] = None
        return result

    # Create output sequence
    out_seq = maskSeq(align, mode=mode, barcode=barcode, barcode_length=barcode_length, barcode_field=barcode_field,
                      primer_field=primer_field, delimiter=delimiter)
    result.results = out_seq
    result.valid = bool(align.error <= max_error) if len(out_seq) > 0 else False

    # Update log with successful alignment results
    result.log['PRIMER'] = align.primer
    result.log['PRORIENT'] = 'RC' if align.rev_primer else 'F'
    result.log['PRSTART'] = align.start
    if 'barcode' in out_seq.annotations:
        result.log['BARCODE'] = out_seq.annotations['barcode']
    if not align.rev_primer:
        align_cut = len(align.align_seq) - align.gaps
        result.log['INSEQ'] = align.align_seq + \
                              str(align.seq.seq[align_cut:])
        result.log['ALIGN'] = align.align_primer
        result.log['OUTSEQ'] = str(out_seq.seq).rjust(len(result.data.seq) + align.gaps)
    else:
        align_cut = len(align.seq) - len(align.align_seq) + align.gaps
        result.log['INSEQ'] = str(align.seq.seq[:align_cut]) + align.align_seq
        result.log['ALIGN'] = align.align_primer.rjust(len(result.data.seq) + align.gaps)
        result.log['OUTSEQ'] = str(out_seq.seq)
    result.log['ERROR'] = align.error

    return result


def maskPrimers(seq_file, primer_file, align_func, align_args={},
                out_file=None, out_args=default_out_args,
                nproc=None, queue_size=None):
    """
    Masks or cuts primers from sample sequences using local alignment

    Arguments:
      seq_file : name of file containing sample sequences.
      primer_file : name of the file containing primer sequences.
      align_func : the function to call for alignment.
      align_arcs : a dictionary of arguments to pass to align_func.
      out_file : output file name. Automatically generated from the input file if None.
      out_args : common output argument dictionary from parseCommonArgs.
      nproc : the number of processQueue processes;
              if None defaults to the number of CPUs.
      queue_size : maximum size of the argument queue;
                   if None defaults to 2*nproc.

    Returns:
      list: a list of successful output file names.
    """
    # Define subcommand label dictionary
    cmd_dict = {alignPrimers: 'align', scorePrimers: 'score', extractPrimers: 'extract'}

    # Print parameter info
    log = OrderedDict()
    log['START'] = 'MaskPrimers'
    log['COMMAND'] = cmd_dict.get(align_func, align_func.__name__)
    log['SEQ_FILE'] = os.path.basename(seq_file)
    if primer_file is not None:
        log['PRIMER_FILE'] = os.path.basename(primer_file)
    if 'mode' in align_args: log['MODE'] = align_args['mode']
    if 'max_error' in align_args: log['MAX_ERROR'] = align_args['max_error']
    if 'start' in align_args: log['START_POS'] = align_args['start']
    if 'length' in align_args: log['LENGTH'] = align_args['length']
    if 'max_len' in align_args: log['MAX_LEN'] = align_args['max_len']
    if 'rev_primer' in align_args: log['REV_PRIMER'] = align_args['rev_primer']
    if 'skip_rc' in align_args: log['SKIP_RC'] = align_args['skip_rc']
    if 'gap_penalty' in align_args:
        log['GAP_PENALTY'] = ', '.join([str(x) for x in align_args['gap_penalty']])
    if 'barcode' in align_args:
        log['BARCODE'] = align_args['barcode']
    if 'barcode' in align_args and align_args['barcode']:
        log['BARCODE_FIELD'] = align_args['barcode_field']
    if 'barcode_length' in align_args and align_args['barcode_length']:
        log['BARCODE_LENGTH'] = align_args['barcode_length']
    log['PRIMER_FIELD'] = align_args['primer_field']
    log['NPROC'] = nproc
    printLog(log)

    # Define alignment arguments and compile primers for align mode
    if primer_file is not None:
        primers = readPrimerFile(primer_file)
        if 'rev_primer' in align_args and align_args['rev_primer']:
            primers = {k: reverseComplement(v) for k, v in primers.items()}
        align_args['primers'] = primers
        align_args['score_dict'] = getDNAScoreDict(mask_score=(0, 1), gap_score=(0, 0))
    if align_func is alignPrimers:
        align_args['primers_regex'] = compilePrimers(primers)
    align_args['delimiter'] = out_args['delimiter']

    # Define feeder function and arguments
    feed_func = feedSeqQueue
    feed_args = {'seq_file': seq_file}
    # Define worker function and arguments
    work_func = processSeqQueue
    work_args = {'process_func': align_func,
                 'process_args': align_args}
    # Define collector function and arguments
    collect_func = collectSeqQueue
    collect_args = {'seq_file': seq_file,
                    'label': 'primers',
                    'out_file': out_file,
                    'out_args': out_args}

    # Call process manager
    result = manageProcesses(feed_func, work_func, collect_func,
                             feed_args, work_args, collect_args,
                             nproc, queue_size)

    # Print log
    result['log']['END'] = 'MaskPrimers'
    printLog(result['log'])

    return result['out_files']


def getArgParser():
    """
    Defines the ArgumentParser

    Returns:
      argparse.ArgumentParser: argument parser object.
    """
    # Define output file names and header fields
    fields = dedent(
             '''
             output files:
                 mask-pass
                     processed reads with successful primer matches.
                 mask-fail
                     raw reads failing primer identification.

             output annotation fields:
                 SEQORIENT
                     the orientation of the output sequence. Either F (input) or RC
                     (reverse complement of input).
                 PRIMER
                     name of the best primer match.
                 BARCODE
                     the sequence preceding the primer match. Only output when the
                     --barcode flag is specified.
             ''')

    # Define ArgumentParser
    parser = ArgumentParser(description=__doc__, epilog=fields,
                            formatter_class=CommonHelpFormatter, add_help=False)
    group_help = parser.add_argument_group('help')
    group_help.add_argument('--version', action='version',
                            version='%(prog)s:' + ' %s %s' %(__version__, __date__))
    group_help.add_argument('-h', '--help', action='help', help='show this help message and exit')
    subparsers = parser.add_subparsers(title='subcommands', metavar='',
                                       help='Alignment method')
    # TODO:  This is a temporary fix for Python issue 9253
    subparsers.required = True

    # Parent parser
    parent_parser = getCommonArgParser(multiproc=True)

    # Align mode argument parser
    parser_align = subparsers.add_parser('align', parents=[parent_parser],
                                         formatter_class=CommonHelpFormatter, add_help=False,
                                         help='Find primer matches using pairwise local alignment.',
                                         description='Find primer matches using pairwise local alignment.')
    group_align = parser_align.add_argument_group('primer alignment arguments')
    group_align.add_argument('-p', action='store', dest='primer_file', required=True,
                              help='A FASTA file containing primer sequences.')
    group_align.add_argument('--maxerror', action='store', dest='max_error', type=float,
                             default=default_primer_max_error, help='Maximum allowable error rate.')
    group_align.add_argument('--maxlen', action='store', dest='max_len', type=int,
                             default=default_primer_max_len,
                             help='''Length of the sequence window to scan for primers.''')
    group_align.add_argument('--gap', nargs=2, action='store', dest='gap_penalty',
                             type=float, default=default_primer_gap_penalty,
                             help='''A list of two positive values defining the gap open
                                   and gap extension penalties for aligning the primers.
                                   Note: the error rate is calculated as the percentage
                                   of mismatches from the primer sequence with gap
                                   penalties reducing the match count accordingly; this may
                                   lead to error rates that differ from strict mismatch
                                   percentage when gaps are present in the alignment.''')
    group_align.add_argument('--revpr', action='store_true', dest='rev_primer',
                              help='''Specify to match the tail-end of the sequence against the
                                   reverse complement of the primers. This also reverses the
                                   behavior of the --maxlen argument, such that the search
                                   window begins at the tail-end of the sequence.''')
    group_align.add_argument('--skiprc', action='store_true', dest='skip_rc',
                              help='Specify to prevent checking of sample reverse complement sequences.')
    group_align.add_argument('--mode', action='store', dest='mode',
                              choices=('cut', 'mask', 'trim', 'tag'), default='mask',
                              help='''Specifies the action to take with the primer sequence.
                                   The "cut" mode will remove both the primer region and
                                   the preceding sequence. The "mask" mode will replace the
                                   primer region with Ns and remove the preceding sequence.
                                   The "trim" mode will remove the region preceding the primer,
                                   but leave the primer region intact. The "tag" mode will
                                   leave the input sequence unmodified.''')
    group_align.add_argument('--barcode', action='store_true', dest='barcode',
                              help='''Specify to annotate reads sequences with barcode sequences
                                   (unique molecular identifiers) found preceding the primer.''')
    group_align.add_argument('--barcodelen', action='store', dest='barcode_length', type=int,
                             help='''Length of the barcode sequence. If not specified, the
                                   barcode sequence is the full sequence preceding the primer.''')
    group_align.add_argument('--bf', action='store', dest='barcode_field', default=default_barcode_field,
                             help='''Name of the barcode annotation field.''')
    group_align.add_argument('--pf', action='store', dest='primer_field', default=default_primer_field,
                             help='''Name of the annotation field containing the primer name.''')
    parser_align.set_defaults(align_func=alignPrimers)

    # Score mode argument parser
    parser_score = subparsers.add_parser('score', parents=[parent_parser],
                                         formatter_class=CommonHelpFormatter, add_help=False,
                                         help='Find primer matches by scoring primers at a fixed position.',
                                         description='Find primer matches by scoring primers at a fixed position.')
    group_score = parser_score.add_argument_group('primer scoring arguments')
    group_score.add_argument('-p', action='store', dest='primer_file', required=True,
                              help='A FASTA file containing primer sequences.')
    group_score.add_argument('--start', action='store', dest='start', type=int, default=default_primer_start,
                             help='The starting position of the primer.')
    group_score.add_argument('--maxerror', action='store', dest='max_error', type=float,
                             default=default_primer_max_error, help='Maximum allowable error rate.')
    group_score.add_argument('--revpr', action='store_true', dest='rev_primer',
                              help='''Specify to match the tail-end of the sequence against the
                                   reverse complement of the primers. This also reverses the
                                   behavior of the --start argument, such that start position is
                                   relative to the tail-end of the sequence.''')
    group_score.add_argument('--mode', action='store', dest='mode',
                              choices=('cut', 'mask', 'trim', 'tag'), default='mask',
                              help='''Specifies the action to take with the primer sequence.
                                   The "cut" mode will remove both the primer region and
                                   the preceding sequence. The "mask" mode will replace the
                                   primer region with Ns and remove the preceding sequence.
                                   The "trim" mode will remove the region preceding the primer,
                                   but leave the primer region intact. The "tag" mode will
                                   leave the input sequence unmodified.''')
    group_score.add_argument('--barcode', action='store_true', dest='barcode',
                              help='''Specify to annotate reads sequences with barcode sequences
                                   (unique molecular identifiers) found preceding the primer.''')
    group_score.add_argument('--barcodelen', action='store', dest='barcode_length', type=int,
                             help='''Length of the barcode sequence. If not specified, the
                                   barcode sequence is the full sequence preceding the primer.''')
    group_score.add_argument('--bf', action='store', dest='barcode_field', default=default_barcode_field,
                             help='''Name of the barcode annotation field.''')
    group_score.add_argument('--pf', action='store', dest='primer_field', default=default_primer_field,
                             help='''Name of the annotation field containing the primer name.''')
    parser_score.set_defaults(align_func=scorePrimers)

    # Extract mode argument parser
    parser_extract = subparsers.add_parser('extract', parents=[parent_parser],
                                           formatter_class=CommonHelpFormatter, add_help=False,
                                           help='Remove and annotate a fixed sequence region.',
                                           description='Remove and annotate a fixed sequence region.')
    group_extract = parser_extract.add_argument_group('region extraction arguments')
    group_extract.add_argument('--start', action='store', dest='start', type=int, default=default_primer_start,
                               help='The starting position of the sequence region to extract.')
    group_extract.add_argument('--len', action='store', dest='length', type=int, required=True,
                               help='The length of the sequence to extract.')
    group_extract.add_argument('--revpr', action='store_true', dest='rev_primer',
                               help='''Specify to extract from the tail end of the sequence with the start
                                    position relative to the end of the sequence.''')
    group_extract.add_argument('--mode', action='store', dest='mode',
                               choices=('cut', 'mask', 'trim', 'tag'), default='mask',
                               help='''Specifies the action to take with the sequence region.
                                    The "cut" mode will remove the region.
                                    The "mask" mode will replace the specified region with Ns.
                                    The "trim" mode will remove the sequence preceding the specified region,
                                    but leave the region intact.
                                    The "tag" mode will leave the input sequence unmodified.''')
    group_extract.add_argument('--barcode', action='store_true', dest='barcode',
                               help='''Specify to remove the sequence preceding the extracted region and
                                    annotate the read with that sequence.''')
    group_extract.add_argument('--barcodelen', action='store', dest='barcode_length', type=int,
                                help='''Length of the barcode sequence. If not specified, the
                                        barcode sequence is the full sequence preceding the primer.''')
    group_extract.add_argument('--bf', action='store', dest='barcode_field', default=default_barcode_field,
                               help='''Name of the barcode annotation field.''')
    group_extract.add_argument('--pf', action='store', dest='primer_field', default=default_primer_field,
                               help='''Name of the annotation field containing the extracted sequence region.''')
    parser_extract.set_defaults(align_func=extractPrimers)

    return parser



if __name__ == '__main__':
    """
    Parses command line arguments and calls main function
    """
    # Parse arguments
    parser = getArgParser()
    checkArgs(parser)
    args = parser.parse_args()
    args_dict = parseCommonArgs(args)

    # Define align_args dictionary to pass to maskPrimers
    if args_dict['align_func'] is alignPrimers:
        args_dict['align_args'] = {'max_error': args_dict['max_error'],
                                   'max_len':args_dict['max_len'],
                                   'rev_primer':args_dict['rev_primer'],
                                   'skip_rc':args_dict['skip_rc'],
                                   'gap_penalty':args_dict['gap_penalty'],
                                   'mode': args_dict['mode'],
                                   'barcode': args_dict['barcode'],
                                   'barcode_length': args_dict['barcode_length'],
                                   'barcode_field': args_dict['barcode_field'],
                                   'primer_field': args_dict['primer_field']}

        del args_dict['max_error']
        del args_dict['max_len']
        del args_dict['rev_primer']
        del args_dict['skip_rc']
        del args_dict['gap_penalty']
        del args_dict['mode']
        del args_dict['barcode']
        del args_dict['barcode_length']
        del args_dict['barcode_field']
        del args_dict['primer_field']
    elif args_dict['align_func'] is scorePrimers:
        args_dict['align_args'] = {'max_error': args_dict['max_error'],
                                   'start':args_dict['start'],
                                   'rev_primer':args_dict['rev_primer'],
                                   'mode': args_dict['mode'],
                                   'barcode': args_dict['barcode'],
                                   'barcode_field': args_dict['barcode_field'],
                                   'barcode_length': args_dict['barcode_length'],
                                   'primer_field': args_dict['primer_field']}
        del args_dict['max_error']
        del args_dict['start']
        del args_dict['rev_primer']
        del args_dict['mode']
        del args_dict['barcode']
        del args_dict['barcode_field']
        del args_dict['barcode_length']
        del args_dict['primer_field']
    elif args_dict['align_func'] is extractPrimers:
        args_dict['primer_file'] = None
        args_dict['align_args'] = {'start':args_dict['start'],
                                   'length':args_dict['length'],
                                   'rev_primer': args_dict['rev_primer'],
                                   'mode':args_dict['mode'],
                                   'barcode': args_dict['barcode'],
                                   'barcode_field': args_dict['barcode_field'],
                                   'barcode_length': args_dict['barcode_length'],
                                   'primer_field': args_dict['primer_field']}
        del args_dict['start']
        del args_dict['length']
        del args_dict['rev_primer']
        del args_dict['mode']
        del args_dict['barcode']
        del args_dict['barcode_length']
        del args_dict['barcode_field']
        del args_dict['primer_field']

    # Call maskPrimers for each sample file
    del args_dict['seq_files']
    if 'out_files' in args_dict:  del args_dict['out_files']
    for i, f in enumerate(args.__dict__['seq_files']):
        args_dict['seq_file'] = f
        args_dict['out_file'] = args.__dict__['out_files'][i] \
            if args.__dict__['out_files'] else None
        maskPrimers(**args_dict)



"""
Unit tests for Annotation module
"""
# Info
__author__ = 'Jason Anthony Vander Heiden'

# Imports
import os
import sys
import time
import unittest
from collections import OrderedDict

# Presto imports
from presto.IO import getFileType, readSeqFile

# Paths
test_path = os.path.dirname(os.path.realpath(__file__))
data_path = os.path.join(test_path, 'data')


class TestIO(unittest.TestCase):
    def setUp(self):
        print('-> %s()' % self._testMethodName)
        # Annotation dictionaries
        self.germ_file = os.path.join(data_path, 'human_igv_trv.fasta')
        self.missing_file = os.path.join(data_path, 'human_igv_trv.bob')

        # Start clock
        self.start = time.time()

    def tearDown(self):
        # End clock
        t = time.time() - self.start
        print('<- %s() %.3f' % (self._testMethodName, t))

    #@unittest.skip('-> getFileType() skipped\n')
    def test_getFileType(self):
        result = getFileType(self.germ_file)
        print(result)
        self.assertEqual('fasta', result)

        result = getFileType(self.missing_file)
        print(result)
        self.assertEqual('fasta', result)

    #@unittest.skip('-> readSeqFile() skipped\n')
    def test_getFileType(self):
        pass
        #result = readSeqFile(self.germ_file)
        #self.assertRaises(SystemExit, readSeqFile(self.missing_file))


if __name__ == '__main__':
    unittest.main()